/**********************************************************************/
/* Macro: rmdmupkg                                                    */
/*        IT Data Mart Unpackage macro                                */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Description:                                                       */
/* This macro is run on an installed image of ITRM in order to        */
/* unpackage all the transport files that were packaged by the        */
/* %rmdmpkg macro.  PROC CIMPORT is used to create the SAS            */
/* data libraries.  This macro may be run in report mode first        */
/* so that only a report of what would be done is produced.           */
/* This is the default.  To actually do the unpackage and create      */
/* the SAS data libraries, specify doPackages=YES.                    */
/*                                                                    */
/* The rmdmupkg macro is typically preceded by an execution of the    */
/* rmdmpkg and rmdmpath macros.                                       */
/*                                                                    */
/* This macro should be executed on the ITRM server tier.             */
/*                                                                    */
/* Note:                                                              */
/* Because some of the packages are initially imported into the WORK  */
/* library, this macro may require large amounts of work space        */
/* to run successfully. The maximum amount of WORK space needed       */
/* is slightly larger than the largest single library to be           */
/* imported.                                                          */
/*                                                                    */
/*--------------------------------------------------------------------*/
/* Parameters:                                                        */
/*     metaserver - Metadata server host name                         */
/*          -REQUIRED-                                                */
/*     metaport - Metadata server port number (usually 8561)          */
/*          -REQUIRED-                                                */
/*     metauser - Userid to access the metadata server                */
/*          -REQUIRED-                                                */
/*     metapass - Password for the metauser userid                    */
/*          -REQUIRED-                                                */
/*     dataMart - Name of the IT datamart that is to be used.         */
/*          -REQUIRED-                                                */
/*     pkgdir -  Input directory containing the files previously      */
/*               created by the %rmdmpkg macro.  This directory       */
/*               must exist prior to running %rmdmupkg and            */
/*               must contain all the CPORT files as well as the      */
/*               package index file.                                  */
/*          -REQUIRED-                                                */
/*     idxname - name of the file that is to be used as an            */
/*               index of the files that are packaged.  This          */
/*               parameter is optional and only needs to be supplied  */
/*               if there are naming conflicts.  The default name     */
/*               is PKGIDX.cpt on Windows and Unix and PKGIDX on z/OS.*/
/*     repos - Name of the repository.  This parameter                */
/*               is optional, the default is FOUNDATION.              */
/*     doReport - Create a report detailing what data libraries       */
/*               are to be unpackaged.  This parameter is optional,   */
/*               valid values are YES or NO, the default is YES.      */
/*     doPackages - Create the SAS data libraries from the CPORT      */
/*               files. This parameter is optional,                   */
/*               valid values are YES or NO, the default is NO.       */
/*     fileDevice - type of device, for example, "SYSDA". This        */
/*               parameter is optional and only used for zOS          */
/*               traditional filesystems.  If not supplied, all       */
/*               new library allocations on zOS will use the value    */
/*               from the FILEDEV SAS system option.                  */
/*     fileUnit - unit of allocation, for example, "CYLS".  This      */
/*               parameter is optional and only used for zOS          */
/*               traditional filesystem.  Valid values are the same   */
/*               values documented for the FILEUNIT SAS option.  If   */
/*               not supplied, all new library allocations on zOS     */
/*               will use the value from the FILEUNIT SAS system      */
/*               option.                                              */
/*     fileSpacePrimary - primary space allocation.  This parameter   */
/*               is optional and only used for zOS traditional        */
/*               filesystems. If not supplied, all new library        */
/*               allocations on zOS will use the value from the       */
/*               FILESPPRI SAS system option.                         */
/*     fileSpaceSecondary - secondary space allocation. This          */
/*               parameter is optional and only used for zOS          */
/*               traditional filesystems. If not supplied, all new    */
/*               library allocations on zOS will use the value from   */
/*               the FILESPSEC SAS system option.                     */
/*     libnameOptions - Addtional options for allocation or any other */
/*               options valid in a zOS LIBNAME statement. This       */
/*               parameter is optional and is only used for zOS       */
/*               traditional filesystems. The default is blank and    */
/*               any supplied values are not validated prior to use.  */
/*     _rc - user-specified return code macro variable.  This         */
/*               parameter is optional.                               */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Error Recovery                                                     */
/* To rerun the macro in the event of an error, you may rerun without */
/* deleting the target libraries in the IT data mart since they       */
/* will be overwritten on subsequent executions of rmdmupkg.sas. This */
/* macro does not update any metadata on the SAS Metadata Server.     */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* History                                                            */
/* 12Jan2012 rmdmupkg version 303000                                  */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Copyright (c) 2009, 2010 SAS Institute, Inc Cary, NC USA           */
/* All rights reserved                                                */
/*                                                                    */
/**********************************************************************/

%macro rmdmupkg (metaServer=,
                metaPort=8561,
                metaUser=,
                metaPass=,
                datamart=,
                pkgdir=,
                idxname=PKGIDX,
                repos=FOUNDATION,
                doReport=YES,
                doPackages=NO,
                fileDevice=,
                fileUnit=,
                fileSpacePrimary=,
                fileSpaceSecondary=,
                libnameOptions=,
                librefFilter=,
                compressAgg=NO,
                smu=YES,
                _rc=);

   %global RMLOGGERLEVEL;
   %let RMLOGGERLEVEL=2;
   %rmcrmsg;

   %global debug;
   %global _exists;
   %local itms_rc;
   %local _pkgscp;
   %local _pkgdelim;
   %local _dataset_parms;
   %local _version;

   %let _version = 303000;
   %let itms_rc = 0;

   %if (&debug ne 1) %then
      %let debug = 0;

   %if &debug %then
      %let _dataset_parms = ;
   %else
      %let _dataset_parms = nolist nowarn;

   %if ("&sysscp" eq "WIN") %then
   %do;
      %let _pkgscp = WIN;
      %let _pkgdelim = %str(\);
   %end;

   %else
   %do;
      %if ("&sysscp" eq "OS") %then
      %do;
         %if %index(%quote(&pkgdir),%str(/)) ne 0 %then
         %do;
            %let _pkgscp = ZFS;
            %let _pkgdelim = %str(/);
         %end;
         %else
         %do;
            %let _pkgscp = OS;
            %let _pkgdelim = %str(.);
         %end;
      %end;

      %else
      %do;
         %let _pkgscp = UNX;
         %let _pkgdelim = %str(/);
      %end;
   %end;

   %if &debug %then
       %put SYSSCP is &sysscp and PKGSCP is &_pkgscp;

   %if &sysver ne 9.2 and &sysver ne 9.3 and &sysver ne 9.4 %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_version_err,
                             noquote
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;


   /* --------------------- Verify Parameters  -----------------------*/

   %if (%quote(&metaServer) eq ) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             metaServer
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (&metaPort eq ) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             metaPort
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&metaUser) eq ) %then
   %do;
       %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             metaUser
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&metaPass) eq ) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             metaPass
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&datamart) eq ) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             datamart
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&pkgdir) eq ) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_required_err,
                             noquote,
                             pkgdir
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;

   %if (%quote(&idxname) eq ) %then
      %let idxname = PKGIDX;
   %if &_pkgscp = OS %then
      %let _idxpath = %qtrim(&pkgdir.)%str(%()&idxname.%str(%));
   %else
      %let _idxpath = %qtrim(&pkgdir.)&_pkgdelim.&idxname..cpt;

   %if (%quote(&repos) eq ) %then
      %let repos = FOUNDATION;

   %if (&compressAgg eq ) %then
      %let compressAgg = NO;

   %if (&doReport eq ) %then
      %let doReport = YES;
   %else
      %let doReport = %upcase(&doReport);
   %if (&doReport ne NO and &doReport ne YES) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_yesno_err,
                             noquote,
                             doReport
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;

   %if (&doPackages eq ) %then
      %let doPackages = NO;
   %else
      %let doPackages = %upcase(&doPackages);
   %if (&doPackages ne NO and &doPackages ne YES and &doPackages ne INDEXONLY) %then
   %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_parm_yesno_err,
                             noquote,
                             doPackages
                             ));
      %let itms_rc=8;
      %goto exit;
   %end;

   %if ("&sysscp" eq "OS") %then
   %do;

      %if (&fileDevice ne ) %then
      %do;
          options filedev="&fileDevice";
      %end;

      %if (&fileUnit ne ) %then
      %do;
          options fileunit=&fileUnit;
      %end;

      %if (&fileSpacePrimary ne ) %then
      %do;
         %let num=0123456789;
         %let chk_num = %verify(&fileSpacePrimary, &num);
         %if &chk_num ne 0 %then
         %do;
            %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                                RMDMUPKG_parm_file_err,
                                noquote,
                                fileSpacePrimary
                                ));
            %let itms_rc=8;
            %goto exit;
         %end;
         %else
         %do;
            options filesppri=&fileSpacePrimary;
         %end;
      %end;

      %if (&fileSpaceSecondary ne ) %then
      %do;
         %let num=0123456789;
         %let chk_num = %verify(&fileSpaceSecondary, &num);
         %if &chk_num ne 0 %then
         %do;
            %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                                RMDMUPKG_parm_file_err,
                                noquote,
                                fileSpaceSecondary
                                ));
            %let itms_rc=8;
            %goto exit;
         %end;
         %else
         %do;
            options filespsec=&fileSpaceSecondary;
         %end;
      %end;

   %end;


   /* --------------------- Macro Definitions  -----------------------*/

   /* Encode the given string to ASCII */
   %macro encodeToAscii(string);
      %if ("&_pkgscp" eq "OS") %then
      %do;
         length encoding $20;
         encoding = getoption("ENCODING");
         &string = kcvt(&string, encoding, "US-ASCII");
      %end;
   %mend encodeToAscii;


   /* Macro to generate global variables for */
   /* dynamically generated names            */
   %macro globalize(var);
      %if %symglobl(&var) = 0 %then
         %global &var;
   %mend globalize;


   /* Generate macro variables for names of members in WORK */
   %macro genWorkDS;

      %global _nummem;

      proc sql;
         create table __pkgm_ as
         select memname, memtype
         from dictionary.members
         where libname="WORK";
      quit;

      data _null_;
         retain count 0;
         length macname mactype $8;

         set __pkgm_ end=last;

         /* Ignore datasets beginning with _PKG */
         if memname ne: "_PKG" and
            memname ne "SASMACR" and
            memname ne "SASGOPT" and
            memname ne "PROFILE" and
            memname ne "REGSTRY" and
            memname ne "FORMATS" then
         do;
            count + 1;
            macname = strip('_mb' !! put(count,5.-L));
            call execute('%globalize(' !! macname !! ');' );
            call symput(macname, memname);
            mactype = strip('_mt' !! put(count,5.-L));
            call execute('%globalize(' !! mactype !! ');' );
            call symput(mactype, memtype);
         end;

         if last then
         do;
            call symput('_nummem', put(count,6.-L));
         end;

      run;

      proc datasets lib=work &_dataset_parms;
         delete __pkgm_;
      run;
      quit;

   %mend genWorkDS;


   /* This macro does a selective clearing out of WORK */
   %macro clearWork;

      %genWorkDS;

      proc datasets lib=work memtype=all &_dataset_parms;
      %do i = 1 %to &_nummem;
            delete &&_mb&i;
      %end;
      run;
      quit;

   %mend clearWork;


   /* Macro to allocate directories for SAS libraries */
   %macro allocdir (fullpath);

      data _null_;
         length path word newpath return_path options $256;
         length _msgtxt $200;

         _msgtxt = "";

         if (not(fileexist("&fullpath")))  then
         do;

            /* Traditional zOS */
            if ("&sysscp" eq "OS" and not index("&fullpath", "/")) then
            do;
               options = "DISP=(NEW,CATLG) " !!
                         strip("&libnameOptions");

               %if &debug %then
               %do;
                  put "Allocdir: zOS options=" options;
               %end;

               newpath = "&fullpath";

               rc = libname("__NEW", newpath, "BASE", options);
               %if &debug %then
               %do;
                  put "Allocdir: libname rc=" rc;
               %end;

               rc = libname("__NEW");

               if (not(fileexist(newpath))) then do;
                  _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                    'sashelp.itms, ' !!
                                    'RMDMUPKG_dir_alloc_1_err, ' !!
                                    'noquote, ' !!
                                    strip(newpath) !!
                                    '))');
                  _error_ = 0;
                  call symput('itms_rc','8');
               end;
               else
               do;
                  _msgtxt = resolve('%rmLogInf(ITMS,%rmmsg(' !!
                                    'sashelp.itms, ' !!
                                    'RMDMUPKG_dir_alloc_note, ' !!
                                    'noquote, ' !!
                                    strip(newpath) !!
                                    '))');
               end;

            end;

            /* Windows */
            else do;
               if ("&sysscp" eq "WIN") then
               do;

                  lvlcount = countw("&fullpath","\");
                  if (substr("&fullpath",1,2) = "\\") then
                  do;
                     path = "\\";
                     unc = 1;
                  end;
                  else
                  do;
                     path = "";
                     unc = 0;
                  end;

                  do i = 1 to lvlcount;
                     word = scan("&fullpath", i, "\");
                     newpath = strip(path) !! strip(word);
                     if i lt lvlcount then
                        newpath = trim(newpath) !! "\";
                     if word ne "" then
                     do;
                        if i = 1 and unc then
                        do;
                           %if &debug %then
                           %do;
                              put "Allocdir: not allocating unc path " newpath;
                           %end;
                        end;
                        else
                        do;
                           if (not(fileexist(newpath))) then
                           do;
                              %if &debug %then
                              %do;
                                 put "Allocdir: allocating " newpath;
                              %end;

                              return_path = dcreate(word, path);

                              %if &debug %then
                              %do;
                                 put "Allocdir: after dcreate " return_path;
                              %end;

                              if return_path = "" then
                              do;
                                 _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                                   'sashelp.itms, ' !!
                                                   'RMDMUPKG_dir_alloc_2_err, ' !!
                                                   'noquote, ' !!
                                                   strip(word) !!
                                                   ',' !!
                                                   strip(path) !!
                                                   '))');
                                 _error_ = 0;
                                 call symput('itms_rc','8');
                              end;
                              else
                              do;
                                 _msgtxt = resolve('%rmLogInf(ITMS,%rmmsg(' !!
                                                   'sashelp.itms, ' !!
                                                   'RMDMUPKG_dir_alloc_note, ' !!
                                                   'noquote, ' !!
                                                   strip(newpath) !!
                                                   '))');
                              end;
                           end;
                           else
                           do;
                              %if &debug %then
                              %do;
                                 put "Allocdir: not allocating " newpath;
                              %end;
                           end;
                        end;

                     end;  /* if word ne blank */

                     path = newpath;

                  end;  /* do loop */

               end;  /* if Windows */

               /* Unix and zFS-based zOS */
               else
               do;

                  lvlcount = countw("&fullpath","/");
                  path = "/";

                  do i = 1 to lvlcount;
                     word = scan("&fullpath", i, "/");
                     newpath = strip(path) !! strip(word);
                     if i lt lvlcount then
                        newpath = trim(newpath) !! "/";
                     if word ne "" then
                     do;

                        if (not(fileexist(newpath))) then
                        do;
                           %if &debug %then
                           %do;
                              put "Allocdir: allocating " newpath;
                           %end;

                           return_path = dcreate(word, path);

                           %if &debug %then
                           %do;
                              put "Allocdir: after dcreate " return_path;
                           %end;

                           if return_path = "" then
                           do;
                              _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                                'sashelp.itms, ' !!
                                                'RMDMUPKG_dir_alloc_2_err, ' !!
                                                'noquote, ' !!
                                                strip(word) !!
                                                ',' !!
                                                strip(path) !!
                                                '))');
                              _error_ = 0;
                              call symput('itms_rc','8');
                           end;
                           else
                           do;
                              _msgtxt = resolve('%rmLogInf(ITMS,%rmmsg(' !!
                                                'sashelp.itms, ' !!
                                                'RMDMUPKG_dir_alloc_note, ' !!
                                                'noquote, ' !!
                                                strip(newpath) !!
                                                '))');
                           end;
                        end;
                        else
                        do;
                           %if &debug %then
                           %do;
                              put "Allocdir: not allocating " newpath;
                           %end;
                        end;

                     end;  /* if word ne blank */

                     path = newpath;

                  end;  /* do loop */

               end;  /* if Unix */

            end;  /* not zOS */

            call symput("_exists","0");

         end;  /* if not fileexist */

         /* already allocated */
         else
            call symput("_exists","1");

      run;

   %mend allocdir;


   /* This macro reformats the summary status datasets for 3.2+ */
   %macro reformat_summarystatus(tblin, tblout, datafmt);

      data &tblout (keep=completehwm ldatm edatm lstpdate);

         format completehwm datetime21.2;
         format ldatm       &datafmt;
         format edatm       &datafmt;
         format lstpdate    datetime21.2;

         %let aggstatus_label_1 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_aggstatus_label_1);
         %let aggstatus_label_2 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_aggstatus_label_2);
         %let aggstatus_label_3 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_aggstatus_label_3);
         %let aggstatus_label_4 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_aggstatus_label_4);

         label completehwm = "&aggstatus_label_1";
         label ldatm = "&aggstatus_label_2";
         label edatm = "&aggstatus_label_3";
         label lstpdate = "&aggstatus_label_4";

         set &tblin;

         ldatm = latestDatetime;
         edatm = earliestDatetime;
         lstpdate = LastUpdateDatetime;
         completehwm = .;

      run;

      /* no longer need old dataset */
      proc datasets lib=work &_dataset_parms;
         delete &tblin;
      run;
      quit;

   %mend reformat_summarystatus;


   /* This macro reformats the summary table datasets for 3.2+ */
   %macro reformat_summarytable(tblname, agecol, agelbl, agefmt);

      %if %upcase(&compressAgg) = YES %then
      %do;
         data &tblname(compress=YES);
      %end;
      %else
      %do;
         data &tblname;
      %end;

         length CompletedDay $1;
         length TimePeriod 8;

         %if %nrquote(&agefmt) ne %then
         %do;
            format TimePeriod &agefmt;
         %end;

         %let aggtable_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_aggtable_label);
         %if %nrquote(&agelbl) ne %then
         %do;
            label TimePeriod = "&agelbl";
         %end;
         label CompletedDay = "&aggtable_label";

         set &tblname;

         %if %nrquote(&agecol) ne %then
         %do;
            TimePeriod = &agecol;
         %end;
         %else
         %do;
            TimePeriod = .;
         %end;
         CompletedDay = "";

      run;

   %mend reformat_summarytable;


   /* This is called by the unpack macro to perform */
   /* special processing for aggregations only.     */
   %macro special_agg_processing(lref, tbl);

      data _null_;
         set _pkgtbl_;
            where (trim(upcase(libref)) = "%trim(%upcase(&lref))" and
                   trim(upcase(tblname)) = "%trim(%upcase(&tbl))");

         /* SummaryStatus role logic */
         if tblrole = "SummaryStatus" then
         do;

            %if &debug %then
            %do;
               put "Calling reformat_summarystatus"
                   " for libref=&lref and"
                   " tblname = &tbl";
            %end;

            if (length(aggtbl) ge 32) then
               aggtbl = substr(aggtbl, 1, 31);
            call execute('%reformat_summarystatus(' !!
                         "&tbl" !!
                         ',' !!
                         '_' !!
                         aggtbl !!
                         ',' !!
                         agingfmt !!
                         ');' );
         end;

         /* SummaryTable role logic */
         if upcase(tblrole) =: "SUMMARYTABLE" or upcase(tblrole) =: "DETAILTABLE" then
         do;

            %if &debug %then
            %do;
               put "Calling reformat_summarytable for "
                 "libref=&lref and tblname = &tbl";
            %end;

            call execute('%reformat_summarytable(' !!
                        "&tbl" !!
                        ',' !!
                        agingcol !!
                        ',' !!
                        aginglbl !!
                        ',' !!
                        agingfmt !!
                        ');' );
         end;

      run;

   %mend special_agg_processing;


   /* This is called by the unpack macro to perform  */
   /* special processing for the ADMIN library only. */
   %macro special_admin_processing(member, type, adminDir);

      %if &debug %then
      %do;
         %put Performing special ADMIN processing on &type &member;
      %end;

      /* Match the SAS dataset with the prefixes we know about */
      %if &type = DATA %then
      %do;

         data _pkgkeys_(keep=clibkey member type);
            %adapter_arrays;
            length prefix $3;
            length suffix $5;
            length clibkey $8;
            length member $32;
            length type $8;

            prefix = "";
            suffix = "";
            rows = dim1(xxxcntrlPrefix);
            cols = dim2(xxxcntrlPrefix);
            l = length("&member");
            if l ge 3 then
               prefix = substr("&member", 1, 3);
            if l ge 6 then
               suffix = trim(substr("&member", 4, l-3));

            if suffix = "CNTRL" or suffix = "DAY" then
            do;
               do i = 1 to rows;
                  do j = 1 to cols;
                     if prefix = xxxcntrlPrefix[i,j] then
                     do;

                        %if &debug %then
                        %do;
                           put "Prefix found for &type &member at " i= j=;
                        %end;

                        clibkey = adapterKeys[i];
                        member = "&member";
                        type = "&type";
                        output _pkgkeys_;
                     end;
                  end;
               end;
            end;

         run;

      %end;

      /* Match the SAS catalog with the prefixes we know about */
      %else
      %do;

         data _pkgkeys_(keep=clibkey member type);
            %adapter_arrays;
            length clibkey $8;
            length member $32;
            length type $8;

            rows = dim1(catalogNames);

            do i = 1 to rows;
               if "&member" = trim(catalogNames[i]) then
               do;

                  %if &debug %then
                  %do;
                     put "Catalog found for &member at " i= ;
                  %end;

                  clibkey = adapterKeys[i];
                  member = "&member";
                  type = "&type";
                  output _pkgkeys_;
               end;
            end;

         run;

      %end;

      /* Merge the results with the chosen library data */
      proc sql;
         create table work._pkgadm_ as
            select c.clibkey, c.clibID, c.clibref, c.clibdir,
                   k.member, k.type
            from work._pkgkeys_ k, work._pkgclib_ c
            where k.clibkey = c.clibkey;
      quit;
      run;

      /* If a admin member is matched with an adapter, then   */
      /* copy it to the appropriate chosen library locations. */
      /* If no match, then copy it into ADMIN.                */
      %let dsida = %sysfunc(open(work._pkgadm_,i));
      %let nobsa = %sysfunc(attrn(&dsida,NOBS));
      %if &nobsa = 0 %then
      %do;

         /* Allocate new a directory for the library */
         /* if necessary                             */
         %allocdir(&adminDir);

         /* copy to admin */
         libname admin "%trim(&adminDir)";
         proc datasets lib=work memtype=&type
                       &_dataset_parms;
         copy out=admin;
         select &member;
         run;
         quit;
         libname admin clear;

         /* this is for reporting only */
         data _pkgspeca_temp(keep=dir exists);
            length dir $256;
            dir = "%trim(&adminDir)";
            exists = input("&_exists",1.);
         run;

      %end;

      %syscall set(dsida);

      %let eofa = 0;
      %do %while(&eofa eq 0);

         %let rca = %sysfunc(fetch(&dsida));

         %if &rca = -1 %then
            %let eofa = 1;

         %else
         %do;

            %if &debug %then
            %do;
               %put reading _pkgadm_;
               %put &member &type &clibkey
                    &clibID &clibref &clibdir;
            %end;

            /* Allocate new a directory for the library */
            /* if necessary                             */
            %allocdir(&clibdir);

            /* copy the member to the chosen library */
            libname &clibref "%trim(&clibdir)";
            proc datasets lib=work memtype=&type
                          &_dataset_parms;
            copy out=&clibref;
            select &member;
            run;
            quit;
            libname &clibref clear;

            /* this is for reporting only */
            data _pkgspeca_temp(keep=dir exists);
               length dir $256;
               dir = "%trim(&clibdir)";
               exists = input("&_exists",1.);
            run;

         %end;  /* else do rc */

      %end;  /* do while */

      %let rca = %sysfunc(close(&dsida));

      /* clean up */
      %if %sysfunc(exist(work._pkgspeca_temp)) %then
      %do;
         proc datasets lib=work &_dataset_parms;
            append base=_pkgspeca_ data=_pkgspeca_temp;
            delete _pkgadm_;
            delete _pkgkeys_;
            delete _pkgspeca_temp;
         run;
         quit;
      %end;
      %else
      %do;
         proc datasets lib=work &_dataset_parms;
            delete _pkgadm_;
            delete _pkgkeys_;
         run;
         quit;
      %end;


   %mend special_admin_processing;


   /* This is called by the unpack macro to perform  */
   /* special processing for the SPIN library only.  */
   %macro special_spin_processing(spinpkg, sdir311);

      %if &debug %then
      %do;
         %put Performing special SPIN processing on &spinpkg.;
      %end;

      data _pkgspin_(keep=clibkey clibref clibdir);
         set _pkgclib_;
            where clibkey = "SMF";
      run;

      /* If we found the SMF adapter, then copy the SPIN      */
      /* library to the appropriate chosen library locations. */
      %let dsids = %sysfunc(open(work._pkgspin_,i));

      %syscall set(dsids);

      %let eofs = 0;
      %do %while(&eofs eq 0);

         %let rcs = %sysfunc(fetch(&dsids));

         %if &rcs = -1 %then
            %let eofs = 1;

         %else
         %do;

            /* Allocate new a directory for the library */
            /* if necessary                             */
            %allocdir(&clibdir);

            %if &doPackages = INDEXONLY %then
            %do;
               libname _lib311 "%trim(&sdir311)";
               libname &clibref "%trim(&clibdir)";
               proc copy in = _lib311
                         out = &clibref;
               run;
               libname _lib311 clear;
               libname &clibref clear;
            %end;

            %else
            %do;
               filename _pkgin "%trim(&spinpkg)";
               libname &clibref "%trim(&clibdir)";
               proc cimport lib = &clibref
                            file = _pkgin;
               run;
               libname &clibref clear;
               filename _pkgin clear;
            %end;

            /* this is for reporting only */
            data _pkgspecs_temp(keep=dir exists);
               length dir $256;
               dir = "%trim(&clibdir)";
               exists = input("&_exists",1.);
            run;

         %end;  /* else do rc */

      %end;  /* do while */

      %let rcs = %sysfunc(close(&dsids));

      /* clean up */
      %if %sysfunc(exist(work._pkgspecs_temp)) %then
      %do;
         proc datasets lib=work &_dataset_parms;
            append base=_pkgspecs_ data=_pkgspecs_temp;
            delete _pkgspin_;
            delete _pkgspecs_temp;
         run;
         quit;
      %end;
      %else
      %do;
         proc datasets lib=work &_dataset_parms;
            delete _pkgspin_;
         run;
         quit;
      %end;

   %mend special_spin_processing;


   /* This driver macro is used to unpackage all libraries */
   %macro unpack();

      /* Clear out WORK before starting */
      %clearWork;

      /* Process all packages listed in the index */
      %let dsid = %sysfunc(open(work._pkglib2,i));
      %let nobs = %sysfunc(attrn(&dsid,NOBS));
      %if &nobs = 0 %then
      %do;
         %rmLogWrn(ITMS, %rmmsg(sashelp.itms,
                                RMDMUPKG_nolib_warn,
                                noquote
                                ));
         %let itms_rc = 4;
         %goto exit_unpack;
      %end;

      %syscall set(dsid);

      %let eof = 0;
      %do %while(&eof eq 0);

         %let rc = %sysfunc(fetch(&dsid));

         %if &rc = -1 %then
            %let eof = 1;

         %else
         %do;

            /* Only unpack matched libraries */
            %if &matched = 1 %then
            %do;

               /* Processing for non-aggregation libraries */
               %if &libType ne 1 %then
               %do;

                  /* ADMIN requires special handling due to 3.2+   */
                  /* using the chosen library for dupchk datasets. */
                  %if &libref = ADMIN %then
                  %do;

                     /* Cimport the package into WORK */
                     %if &doPackages = INDEXONLY %then
                     %do;
                        libname _lib311 "%trim(&dir311)";
                        proc copy in = _lib311
                                  out = work;
                             exclude sasmacr/mt=catalog;
                        run;
                        libname _lib311 clear;
                     %end;

                     %else
                     %do;
                        filename _pkgin "%trim(&file32)";
                        proc cimport lib = work
                                     file = _pkgin;
                        run;
                        filename _pkgin clear;
                     %end;

                     /* Get the list of chosen libraries for  */
                     /* the current IT data mart.             */
                     %get_chosen_libs(ADMIN, _pkgdm_);

                     /* Get the names of the members in WORK */
                     %genWorkDS;

                     /* Do special admin processing for each member */
                     %do i = 1 %to &_nummem;
                        %special_admin_processing(&&_mb&i, &&_mt&i, &dir32);
                     %end;

                     /* If ADMIN never was created, then go ahead  */
                     /* and create it just so the datamart has one */
                     %let newdir = &dir32;  /* S0672617 */
                     %if (%sysfunc(fileexist("&newdir")) = 0) %then
                     %do;
                        %allocdir(&newdir);
                     %end;


                     /* no longer need this dataset */
                     proc datasets lib=work &_dataset_parms;
                        delete _pkgclib_;
                     run;
                     quit;

                     /* Clear out WORK for the next package */
                     %clearWork;

                  %end;  /* if libref = ADMIN */

                  %else
                  %do;

                     /* SPIN requires special handling also.   */
                     %if &libref = SPIN %then
                     %do;

                        /* Get the list of chosen libraries for  */
                        /* the current IT data mart.             */
                        %get_chosen_libs(SPIN, _pkgdm_);

                        /* Do special spin processing */
                        %special_spin_processing(&file32,%trim(&dir311));

                        /* If SPIN never was created, then go ahead   */
                        /* and create it just so the datamart has one */
                        %let newdir = &dir32;  /* S0672617 */
                        %if (%sysfunc(fileexist("&newdir")) = 0) %then
                        %do;
                           %allocdir(&newdir);
                        %end;

                        /* no longer need this dataset */
                        proc datasets lib=work &_dataset_parms;
                           delete _pkgclib_;
                        run;
                        quit;

                     %end;  /* if libref = SPIN */

                     %else
                     %do;

                        /* This section handles all other non-aggregation */
                        /* libraries like staging libraries. Note that    */
                        /* this will also cimport any SAS catalogs        */
                        /* (except for SASMACR which is excluded in the   */
                        /* rmdmpkg macro).                                */

                        /* Allocate new a directory for the library */
                        /* if necessary                             */
                        %allocdir(&dir32);

                        /* Cimport the package to the final destination */
                        %if &doPackages = INDEXONLY %then
                        %do;
                           libname _lib311 "%trim(&dir311)";
                           libname &libref "%trim(&dir32)";
                           proc copy in = _lib311
                                     out = &libref;
                           run;
                           libname _lib311 clear;
                           libname &libref clear;
                        %end;

                        %else
                        %do;

                           /* Don't cimport staged libraries if the */
                           /* stagedLibs=NO option was set.         */
                           %if &libType ne 3 %then
                           %do;
                              filename _pkgin "%trim(&file32)";
                              libname &libref "%trim(&dir32)";
                              proc cimport lib = &libref
                                           file = _pkgin;
                              run;
                              libname &libref clear;
                              filename _pkgin clear;
                           %end;

                        %end;

                     %end;  /* if not spin */

                  %end;  /* if not admin */

                  /* Clear out WORK for the next package */
                  %clearWork;

               %end;  /* if libType ne 1 */

               /* Aggregation libraries require special processing. */
               /* Only SAS datasets (not catalogs) are cimported.   */
               %else
               %do;

                  /* Cimport the package into WORK */
                  %if &doPackages = INDEXONLY %then
                  %do;
                     libname _lib311 "%trim(&dir311)";
                     proc copy in = _lib311
                               out = work;
                     run;
                     libname _lib311 clear;
                  %end;

                  %else
                  %do;
                     filename _pkgin "%trim(&file32)";
                     proc cimport lib = work
                                  file = _pkgin
                                  memtype = ds;
                     run;
                     filename _pkgin clear;
                  %end;

                  /* Get the names of the datasets in WORK */
                  %genWorkDS;

                  /* Perform special agg processing for each dataset */
                  %do i = 1 %to &_nummem;
                     %special_agg_processing(&libref, &&_mb&i);
                  %end;

                  /* Allocate new a directory for the library */
                  /* if necessary                             */
                  %allocdir(&dir32);

                  %genWorkDS;

                  /* Write the tables from work to the new library */
                  %if &_nummem ge 1 %then
                  %do;

                     libname &libref "%trim(&dir32)";
                     proc datasets lib=work memtype=data
                                   &_dataset_parms;
                     copy out=&libref;
                     %do i = 1 %to &_nummem;
                        select &&_mb&i;
                     %end;
                     run;
                     quit;
                     libname &libref clear;

                  %end;

                  /* Clear out WORK for the next package */
                  %clearWork;

               %end;  /* if agglib */

            %end;  /* if matched */

         %end;  /* else do rc */

      %end;  /* do while */

      %let rc = %sysfunc(close(&dsid));

   %exit_unpack:
   %mend unpack;


   /* This macro is included by other macros so      */
   /* that they all use the same array definitions.  */
   /* All adapter arrays should have the same        */
   /* number of row elements, each row corresponding */
   /* to a different adapter.                        */
   %macro adapter_arrays;

      /* Adapter keys that map to the other arrays. */
      array adapterKeys [24] $ 8 _temporary_
         (
          "HPOVPA",
          "HPOVREP",
          "AS400",
          "DCOLLECT",
          "EREP",
          "IMF",
          "IMS",
          "NTSMFMXG",
          "SMF",
          "TMON2CIC",
          "TMONDB2",
          "TMS",
          "TPF",
          "VMMON",
          "NTSMF",
          "PATROL",
          "SAPR3",
          "SAR",
          "SCOM",
          "SNMP",
          "VMWARE",
          "VMWAREDA",
          "WEBLOG",
          "UW"
         );


      /* Catalog names that correspond to the adapterKeys array. */
      /* This is used for copying duplicate checking catalogs.   */
      array catalogNames [24] $ 8 _temporary_
         (
          "HPOVPA",
          "HPOVREP",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "NTSMF",
          "PATROL",
          "SAP",
          "SAR",
          "SCOM",
          "SNMP",
          "VMWARE",
          "",
          "WEBLOG",
          ""
         );

      /* xxxCNTRL names that correspond to the adapterKeys array. */
      /* This is used for copying duplicate checking datasets.    */
      array xxxcntrlPrefix [24,12] $ 3 _temporary_
         (
          "HPA",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "APP","CPU","DIS","FIL","GLO","LSY","LVO","NET","PRO","TRA",   "",   "",
             "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "DCO",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "ERP",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "IMF",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
             "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "NTS",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "SMF",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "TM2",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "TMD",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "TMS",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "TPF",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "VMM",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "NTS",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "PAT",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "SAP","BAT","WHM","RCO","RCD","RSO","RSD","RSS","SPA","SPP","TAB",   "",
          "SAR",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "SCO",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "SNM",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "VHS","VGS","VHD","VGD","VHN","VGN","VHG","VGG","VHC","VGC","VTA","VEV",
             "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
          "WWW",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   "",
             "",   "",   "",   "",   "",   "",   "",   "",   "",   "",   ""    ""
         );


      /* Transformation classes that correspond to the adapterKeys array. */
      /* This is used verification purposes in the get_chosen_libs macro. */
      array xformClasses [24] $ 100 _temporary_
         (
          "com.sas.solutions.itms.models.transformation.hpovpa.HPOVPATransformationType",
          "com.sas.solutions.itms.models.transformation.hpovrep.HPOVREPTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.as400.MXGAS400TransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.dcollect.MXGDCOLLECTTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.erep.MXGEREPTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.imf.MXGIMFTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.ims.MXGIMSTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.ntsmfmxg.MXGNTSMFMXGTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.smf.MXGSMFTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.tmon2cic.MXGTMON2CICTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.tmondb2.MXGTMONDB2TransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.tms.MXGTMSTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.tpf.MXGTPFTransformationType",
          "com.sas.solutions.itms.models.transformation.mxg.vmmon.MXGVMMONTransformationType",
          "com.sas.solutions.itms.models.transformation.ntsmf.NTSMFTransformationType",
          "com.sas.solutions.itms.models.transformation.patrol.PATROLTransformationType",
          "com.sas.solutions.itms.models.transformation.sapr3.SAPR3TransformationType",
          "com.sas.solutions.itms.models.transformation.sar.SARTransformationType",
          "com.sas.solutions.itms.models.transformation.scom.SCOMTransformationType",
          "com.sas.solutions.itms.models.transformation.snmp.SNMPTransformationType",
          "com.sas.solutions.itms.models.transformation.vmware.VMWARETransformationType",
          "com.sas.solutions.itms.models.transformation.vmwareda.VMWAREDATransformationType",
          "com.sas.solutions.itms.models.transformation.weblog.WEBLOGTransformationType",
          "com.sas.solutions.itms.models.transformation.UserWrittenStagingTransformationType"
         );

   %mend adapter_arrays;


   /* This macro will determine all chosen libraries for an  */
   /* IT data mart.  These libraries are a destination for   */
   /* the duplicate checking datasets.  The macro can also   */
   /* return all designated spin libraries, since the        */
   /* retrieval of those libraries is a similar process.     */
   %macro get_chosen_libs(library, itdmds);

      data _pkgclib_(keep=clibkey clibID clibref clibdir);

         length parentURI treeURI jobURI actURI stepURI propsetURI
                propURI xformURI targURI libURI dirURI $256;
         length jobName propSetName propName $60;
         length className $100;
         length jobid $17;
         length adapter $8;
         length clibkey $8;
         length clibID $17;
         length clibref $8;
         length clibdir $256;

         length oldestName $60;
         length oldestID $17;
         length oldestRef $8;
         length oldestDir $256;
         length oldestMCchar $18;
         length tempMCchar $18;
         length tempName $60;
         length _msgtxt $200;

         parentURI = "";
         treeURI = "";
         jobURI = "";
         actURI = "";
         stepURI = "";
         propsetURI = "";
         propURI = "";
         xformURI = "";
         targURI = "";
         libURI = "";
         dirURI = "";
         jobName = "";
         propSetName = "";
         propName = "";
         className = "";
         jobid = "";
         adapter = "";
         clibkey = "";
         clibID = "";
         clibref = "";
         clibdir = "";

         currentAdapter = 0;
         oldestID = "";
         oldestRef = "";
         oldestDir = "";

         /* set to the "latest" time for comparison purposes */
         oldestMC = datetime();
         oldestMCchar = put(oldestMC, datetime18.);

         /* set to a high value for comparison purposes */
         oldestName = "ZZZZZZZZ";
         %encodeToAscii(oldestName);

         tempMC = 0;
         tempMCchar = "";
         tempName = "";
         _msgtxt = "";


         /* arrays to handle the recursion */
         array parents [1000] $ 256 _temporary_;
         array trees [1000] $ 256 _temporary_;
         array treeCount [1000] _temporary_;

         %adapter_arrays;

         numAdapters = dim(adapterKeys);

         set &itdmds;

         rc = metadata_getnasn(itdmURI, "Trees", 1, treeURI);
         if (treeURI eq "") then
         do;
            _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                      'sashelp.itms, ' !!
                      'RMDMUPKG_tree_err, ' !!
                      'noquote ' !!
                      '))');
            call symput('itms_rc','8');
            stop;
         end;

         parentURI = treeURI;

         /* start recursive routine */
         level = 0;
         link getJobs;

         return;

         /* recursive section - traverse all trees */
         getJobs:

            /* get the subtrees from a given parent tree */
            tree = 0;
            do until (treeURI eq "");

               tree + 1;
               treeURI  = "";

               %if &debug %then
               %do;
                  put 'Before getnasn of subtrees ' parentURI=;
               %end;

               rc = metadata_getnasn(parentURI, "SubTrees",
                                     tree, treeURI);

                %if &debug %then
                %do;
                   put 'After getnasn of subtrees ' treeURI= rc=;
               %end;

               if (treeURI ne "") then
               do;

                  /* get members of the tree and look for jobs */
                  mem = 0;
                  do until (jobURI = "");

                     mem + 1;
                     jobURI = "";
                     rc = metadata_getnasn(treeURI, "Members",
                                           mem, jobURI);

                     %if &debug %then
                     %do;
                        put 'After getnasn of tree members ' jobURI= rc=;
                     %end;

                     if (jobURI ne "") then
                     do;

                        /* Is this member a job? */
                        if (scan(jobURI, 2, ":\") eq "Job") then
                        do;

                           rc = metadata_getattr(jobURI, "Name", jobName);
                           rc = metadata_getattr(jobURI, "ID", jobid);

                           %if &debug %then
                           %do;
                              put ' ';
                              put 'Found a job ' jobName= jobid=;
                           %end;

                           /* look for activities */
                           act = 0;
                           do until (actURI = "");

                              act + 1;
                              actURI = "";
                              rc = metadata_getnasn(jobURI, "JobActivities",
                                                    act, actURI);

                              %if &debug %then
                              %do;
                                 put 'After getnasn of JobActivities ' actURI= rc=;
                              %end;

                              if (actURI ne "") then
                              do;

                                 if (scan(actURI, 2, ":\") eq "TransformationActivity") then
                                 do;

                                    /* look for steps */
                                    step = 0;
                                    do until (stepURI = "");

                                       step + 1;
                                       stepURI = "";
                                       rc = metadata_getnasn(actURI, "Steps",
                                                             step, stepURI);

                                       %if &debug %then
                                       %do;
                                          put 'After getnasn of Steps ' stepURI= rc=;
                                       %end;

                                       if (stepURI ne "") then
                                       do;

                                          if (scan(stepURI, 2, ":\") eq "TransformationStep") then
                                          do;

                                             /* look for a staging propertyset to confirm that */
                                             /* this step is a staging transformation.         */
                                             propset = 0;
                                             stagingXformFound = 0;
                                             do until (propsetURI = "" or stagingXformFound = 1);

                                                propset + 1;
                                                propsetURI = "";
                                                rc = metadata_getnasn(stepURI, "PropertySets",
                                                                      propset, propsetURI);

                                                %if &debug %then
                                                %do;
                                                   put 'After getnasn of PropertySets ' propsetURI= rc=;
                                                %end;

                                                if (propsetURI ne "") then
                                                do;

                                                   if (scan(propsetURI, 2, ":\") eq "PropertySet") then
                                                   do;

                                                      rc = metadata_getattr(propsetURI, "Name", propSetName);
                                                      if propSetName = 'STAGING' then
                                                      do;

                                                         %if &debug %then
                                                         %do;
                                                            put 'Found a staging xform ...';
                                                         %end;

                                                         StagingXformFound = 1;

                                                         /* Validate class name for staging xform */
                                                         /* and set clibkey.                      */
                                                         prop = 0;
                                                         do until (propURI = "");

                                                            prop + 1;
                                                            propURI = "";
                                                            rc = metadata_getnasn(stepURI, "Properties",
                                                                                  prop, propURI);

                                                            %if &debug %then
                                                            %do;
                                                               put 'After getnasn of Properties ' propURI= rc=;
                                                            %end;

                                                            if (propURI ne "") then
                                                            do;

                                                               if (scan(propURI, 2, ":\") eq "Property") then
                                                               do;

                                                                  rc = metadata_getattr(propURI, "Name", propName);
                                                                  if upcase(propName) = 'CLASS' then
                                                                  do;

                                                                     rc = metadata_getattr(propURI, "DefaultValue", className);
                                                                     currentAdapter = 0;

                                                                     do i = 1 to numAdapters;
                                                                        if className = xformClasses[i] then
                                                                        do;
                                                                           currentAdapter = i;
                                                                           leave;
                                                                        end;
                                                                     end;

                                                                     if currentAdapter = 0 then
                                                                     do;
                                                                        _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                                                                  'sashelp.itms, ' !!
                                                                                  'RMDMUPKG_unk_staging_xform_err, ' !!
                                                                                  'noquote, ' !!
                                                                                  strip(classname) !!
                                                                                  '))');
                                                                     end;

                                                                    %if &debug %then
                                                                     %do;
                                                                        put "adapter=" adapterKeys[currentAdapter];
                                                                     %end;

                                                                  end;  /* if prop name = CLASS */

                                                               end;  /* if property */

                                                            end;  /* if - non blank PropURI */

                                                         end;  /* do until - no more properties */

                                                         /* If valid staging xform, then continue on */
                                                         if currentAdapter then
                                                         do;

                                                            /* get the classifier map */
                                                            xform = 0;
                                                            do until (xformURI = "");

                                                               xform + 1;
                                                               xformURI = "";

                                                               rc = metadata_getnasn(stepURI, "Transformations",
                                                                                     xform, xformURI);

                                                               %if &debug %then
                                                               %do;
                                                                  put 'After getnasn of xforms ' xformURI= rc=;
                                                               %end;

                                                               if (xformURI ne "") then
                                                               do;

                                                                  /* ADMIN logic - look for target tables */
                                                                  /* to get the chosen libraries          */
                                                                  if "&library" = "ADMIN" then
                                                                  do;

                                                                     /* We have a classifier map for the   */
                                                                     /* xform. Look for target tables.     */
                                                                     targ = 0;
                                                                     do until (targURI = "");

                                                                        targ + 1;
                                                                        targURI = "";

                                                                        rc = metadata_getnasn(xformURI, "ClassifierTargets",
                                                                                        targ, targURI);

                                                                        %if &debug %then
                                                                        %do;
                                                                           put 'After getnasn of targets ' targURI= rc=;
                                                                        %end;

                                                                        if (targURI ne "") then
                                                                        do;

                                                                           /* Got a classifier target, now get the library */
                                                                           if (scan(targURI, 2, ":\") eq "PhysicalTable") then
                                                                           do;

                                                                              rc = metadata_getnasn(targURI, "TablePackage",
                                                                                                    1, libURI);

                                                                              rc = metadata_getattr(libURI, "MetadataCreated", tempMCchar );
                                                                              tempMC = input(tempMCchar, datetime18.);

                                                                              %if &debug %then
                                                                              %do;
                                                                                 put 'After getnasn of library ' libURI= rc= tempMCchar= oldestMCchar=;
                                                                              %end;

                                                                              /* See if this is the oldest library.  */
                                                                              /* If so, then this is the chosen one. */
                                                                              setOldest = 0;
                                                                              if tempMC lt oldestMC then
                                                                              do;
                                                                                 setOldest = 1;
                                                                              end;

                                                                              /* If the timestamps are equal then we need to do */
                                                                              /* an additional check on name.  This check must  */
                                                                              /* done in ASCII because the result we get here   */
                                                                              /* must be consistent with the staging jobs which */
                                                                              /* are always generated on Windows via Java.      */
                                                                              /* See S0665721.                                  */
                                                                              if tempMC = oldestMC then
                                                                              do;
                                                                                 rc = metadata_getattr(libURI, "Name", tempName);
                                                                                 tempName = upcase(compress(tempName,,'adk'));
                                                                                 %encodeToAscii(tempName);
                                                                                 if tempName le oldestName then
                                                                                    setOldest = 1;

                                                                                 %if &debug %then
                                                                                 %do;
                                                                                    put 'After name comp check: ' tempName= oldestName= setOldest=;
                                                                                 %end;

                                                                              end;

                                                                              /* If setOldest is 1 then we have a potential chosen */
                                                                              /* one.  Set the oldest variables to the current lib.*/
                                                                              if setOldest = 1 then
                                                                              do;
                                                                                 oldestMC = tempMC;
                                                                                 oldestMCchar = put(oldestMC, datetime18.);
                                                                                 rc = metadata_getattr(libURI, "Name", tempName);
                                                                                 oldestName = upcase(compress(tempName,,'adk'));
                                                                                 %encodeToAscii(oldestName);
                                                                                 rc = metadata_getattr(libURI, "ID", oldestID);
                                                                                 rc = metadata_getattr(libURI, "Libref", oldestRef);
                                                                                 rc = metadata_getnasn(libURI, 'UsingPackages',
                                                                                                       1, dirURI);
                                                                                 rc = metadata_getattr(dirURI,"DirectoryName", oldestDir);
                                                                                 setOldest = 0;

                                                                                 %if &debug %then
                                                                                 %do;
                                                                                    put "***setting: " oldestName= oldestID= oldestRef= oldestDir= tempMCchar=;
                                                                                 %end;

                                                                              end;  /* if setOldest = 1 */

                                                                           end;  /* if PhysicalTable */

                                                                        end; /* if - non blank target table */

                                                                     end;  /* do until - no more target tables */

                                                                     /* can write out chosen info now */
                                                                     clibkey = adapterKeys[currentAdapter];
                                                                     clibID = oldestID;
                                                                     clibref = oldestRef;
                                                                     clibDir = oldestDir;

                                                                     output _pkgclib_;

                                                                     /* reset values for the next classifier map */
                                                                     oldestName = "ZZZZZZZZ";
                                                                     %encodeToAscii(oldestName);
                                                                     oldestID = "";
                                                                     oldestRef = "";
                                                                     oldestDir = "";
                                                                     oldestMC = datetime();
                                                                     oldestMCchar = put(oldestMC, datetime18.);

                                                                  end;  /* if library = ADMIN */

                                                                  /* SPIN logic, look for the SPIN */
                                                                  /* library specification         */
                                                                  else if "&library" = "SPIN" then
                                                                  do;

                                                                     /* We have a classifier map for the   */
                                                                     /* xform. Look for source specs.      */
                                                                     lib = 0;
                                                                     do until (libURI = "");

                                                                        lib + 1;
                                                                        libURI = "";

                                                                        rc = metadata_getnasn(xformURI, "SourceSpecifications",
                                                                                        lib, libURI);

                                                                        %if &debug %then
                                                                        %do;
                                                                           put 'After getnasn of specs ' libURI= rc=;
                                                                        %end;

                                                                        if (libURI ne "") then
                                                                        do;

                                                                           /* Reusing the "oldest" variables for convenience      */
                                                                           /* even though no date-time checking is done for SPIN. */
                                                                           rc = metadata_getattr(libURI, "ID", oldestID);
                                                                           rc = metadata_getattr(libURI, "Libref", oldestRef);
                                                                           rc = metadata_getnasn(libURI, 'UsingPackages',
                                                                                                 1, dirURI);

                                                                           /* make sure that its a directory */
                                                                           if (scan(dirURI, 2, ":\") eq "Directory") then
                                                                           do;
                                                                              rc = metadata_getattr(dirURI,"DirectoryName", oldestDir);

                                                                              clibkey = adapterKeys[currentAdapter];
                                                                              clibID = oldestID;
                                                                              clibref = oldestRef;
                                                                              clibDir = oldestDir;

                                                                              output _pkgclib_;

                                                                              oldestID = "";
                                                                              oldestRef = "";
                                                                              oldestDir = "";
                                                                           end;

                                                                        end; /* if - non blank library */

                                                                     end;  /* do until - no more libraries */

                                                                  end;  /* if library = SPIN */

                                                               end; /* if - non blank xform */

                                                            end;  /* do until - no more xforms */

                                                         end;  /* if currentAdapter */

                                                      end;  /* if propSetname = STAGING */

                                                   end;  /* if PropertySet */

                                                end;  /* if - non blank PropertySet */

                                             end;  /* do until - no more PropertySets */

                                          end;  /* if transformationStep */

                                       end;  /* if - non blank step */

                                    end;  /* do until - no more steps */

                                 end; /* if TransformationActivity */

                              end;  /* if - non blank activity */

                           end; /* do until - no more activities */

                        end;  /* if job */

                     end; /* if - non blank member uri */

                  end; /* do until - no more members */

                  /* save current level information in arrays */
                  /* before recursing                         */
                  level + 1;
                  parents[level] = parentURI;
                  trees[level] = treeURI;
                  treeCount[level] = tree;
                  parentURI = treeURI;

                  link getJobs;

                  /* restore current level information from */
                  /* arrays after recursing                 */
                  treeURI = trees[level];
                  parentURI = parents[level];
                  tree = treeCount[level];
                  level = level - 1;

               end;  /* if - non blank sub tree uri */

            end;  /* do until - no more subtrees */

         return;

      run;

      %if &debug %then
      %do;
         title 'Work._pgkclib_ Contents';
         proc print data = work._pkgclib_;
         run;
      %end;


   %mend get_chosen_libs;

   /*----------------- End of Macro Definitions  --------------------*/


   /*------------------------  Setup --------------------------------*/

   options metaserver="&metaserver"
           metaport=&metaport
           metauser="&metauser"
           metapass="&metapass"
           metaRepository="&repos";

   options nofmterr;


   /* ------------------- Validate Input Directory -------------------*/

   data _null_;

      length _msgtxt $200;
      _msgtxt = "";

      if (not(fileexist("&pkgdir"))) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_exist_1_err, ' !!
                           'noquote, ' !!
                           "&pkgdir" !!
                           '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_exist_2_err, ' !!
                           'noquote ' !!
                           '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_exist_3_err, ' !!
                           'noquote ' !!
                           '))');
         call symput('itms_rc','8');
         stop;
      end;

      rc = filename("_pkgin", "&pkgdir");
      if (rc) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_alloc_1_err, ' !!
                           'noquote, ' !!
                           "&pkgdir" !!
                           '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_alloc_2_err, ' !!
                           'noquote ' !!
                           '))');
         call symput('itms_rc','8');
         stop;
      end;

      /* Prove that its a directory */
      dirid = dopen('_pkgin');
      if (not dirid) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_dir_1_err, ' !!
                           'noquote, ' !!
                           "&pkgdir" !!
                           '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_dir_2_err, ' !!
                           'noquote ' !!
                           '))');
         call symput('itms_rc','8');
         stop;
      end;

      /* Make sure the package index file exists. */
      %if &_pkgscp = OS %then
         %let fname = &idxname;
      %else
         %let fname = &idxname..cpt;
      rc = mopen(dirid, "&fname","i");
      if (not rc) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_file_1_err, ' !!
                           'noquote, ' !!
                           "&fname" !!
                           '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_file_2_err, ' !!
                           'noquote ' !!
                           '))');
         call symput('itms_rc','8');
         stop;
      end;

      rc = dclose(dirid);
      if (rc) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                           'sashelp.itms, ' !!
                           'RMDMUPKG_pkgdir_close_err, ' !!
                           'noquote, ' !!
                           "&pkgdir" !!
                           '))');
         call symput('itms_rc','8');
         stop;
      end;

      rc = filename("_pkgin", "");

   run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;


   /*------------------- Validate IT Datamart -----------------------*/

   data _pkgdm_(keep=itdmURI);

      length searchString $256;
      length itdmURI $256;
      length _msgtxt $200;

      itdmURI = "";
      _msgtxt = "";

      searchString = "omsobj:Group?" !!
                     "Group[@Name eq '&datamart' and " !!
                     "@PublicType eq 'ITDataMart']";

      /* get the start of the data mart tree */
      nobj = metadata_getnobj(searchString, 1, itdmURI);

      if (nobj eq -1) then
      do;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                      'sashelp.itms, ' !!
                      'RMDMUPKG_mdserver_conn_err, ' !!
                      'noquote ' !!
                      '))');
         call symput('itms_rc','8');
         stop;
      end;

      if (itdmURI eq "") then
      do;
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                      'sashelp.itms, ' !!
                      'RMDMUPKG_datamart_1_err, ' !!
                      'noquote, ' !!
                      "&datamart" !!
                      '))');
         _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                      'sashelp.itms, ' !!
                      'RMDMUPKG_datamart_2_err, ' !!
                      'noquote ' !!
                      '))');
         call symput('itms_rc','8');
         stop;
      end;
      else
      do;
         output _pkgdm_;
      end;

      %if &debug %then
      %do;
         put 'After getnobj of datamart ' itdmURI= nobj=;
      %end;

    run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;


   /* -------------------- Unpackage Index File ----------------------*/

   filename _pkgin "&_idxpath";

   /* This will import both _pkglib_ and _pkgtbl_ */
   proc cimport lib=work infile=_pkgin;
   run;

   filename _pkgin clear;

   %if &debug %then
   %do;
      title 'Work._pgklib_ Contents';
      proc print data = work._pkglib_;
      run;

      title 'Work._pgktbl_ Contents';
      proc print data = work._pkgtbl_;
      run;
   %end;

   /* build an index on _pkgtbl_ */
   proc sort data = _pkgtbl_;
      by libref tblname;
   run;

   proc datasets lib = work &_dataset_parms;
      modify _pkgtbl_;
      index create keyvars = (libref tblname) / nomiss unique;
   run;
   quit;


   /* -------- Validate libraries and retrieve directory info --------*/

   data _pkglib2(keep=name libid libref dir311 dir32 file311 filename
                      file32 libType matched alreadyExists);

      length libURI dirURI $256;
      length file32 dir32 $256;
      length _msgtxt $200;
      length ver_src $6;

      libURI = "";
      dirURI = "";
      dir32 = "";
      file32 = "";
      matched = 0;
      alreadyExists = 0;
      _msgtxt = "";
      ver_src = "";

      %if &librefFilter ne %then
      %do;
         set work._pkglib_;
            where upcase(libref) = "%upcase(&librefFilter)";
      %end;
      %else
      %do;
         set work._pkglib_;
      %end;

      /* Validate that the package was created under the same    */
      /* macro version of rmdmpkg.                               */
      if _n_ = 1 then
      do;
         if ver_src ne "&_version" then do;
            _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                              'sashelp.itms, ' !!
                              'RMDMUPKG_version_mismatch_1, ' !!
                              'noquote ' !!
                              '))');
            _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                              'sashelp.itms, ' !!
                              'RMDMUPKG_version_mismatch_2, ' !!
                              'noquote ' !!
                              '))');
            call symput('itms_rc','8');
            stop;
         end;
      end;


      /* Only process libraries that were successfully found and */
      /* packaged on the source system side.                     */
      if found = 1 then
      do;

         /* Build input cport file name.  If no filename, we use */
         /* the file311 field which may contain a message, e.g., */
         /* the stagedLibs=NO option.                            */
         if filename ne "" then do;
            %if &_pkgscp eq OS %then %do;
               file32 = "&pkgdir" !! "(" !!
                        trim(strip(filename))
                        !! ")";
               %end;
            %else %do;
               file32 = "&pkgdir" !! "&_pkgdelim" !!
                      strip(filename);
               %end;
            end;


         else
            file32 = strip(file311);
         /* try to find the same-named library in the metadata */
         searchString = "omsobj:SASLibrary?SASLibrary[" !!
                        "@Name eq '" !! strip(name) !!
                        "' and @Libref eq '" !! strip(libref) !! "']";
         i = 1;
         nobj = 1;
         do while (nobj > 0);

            nobj = metadata_getnobj(searchString, i, libURI);

            %if &debug %then
            %do;
               put "Looking for library in metadata "
                   name= libref= libid= nobj= libURI= ;
            %end;

            /* If the target is a datamart from SMU (which is the normal  */
            /* operation), then validate that the ids match.  Otherwise,  */
            /* bypass the id check and assume that the name and libref    */
            /* give us enough uniqueness to go with.                      */
            if (libURI ne "") then
            do;
               %if (&smu = YES) %then
               %do;
                  reposid = scan(libURI, 3, ":\");
                  if libid = reposid then
                     matched = 1;
               %end;
               %else
               %do;
                  matched = 1;
               %end;
            end;

            i + 1;
            if matched = 1 then
               nobj = -99;

         end;

         if matched = 1 then
         do;

            /* retrieve the directory location for the library */
            rc = metadata_getnasn(libURI, 'UsingPackages',
                 1, dirURI);

            if (rc le 0) then
            do;
               if rc = -1 then
               do;
                  _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                    'sashelp.itms, ' !!
                                    'RMDMUPKG_mdserver_conn_err, ' !!
                                    'noquote ' !!
                                    '))');
               end;
               else
               do;
                  _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                                    'sashelp.itms, ' !!
                                    'RMDMUPKG_usingpkg_err, ' !!
                                    'noquote, ' !!
                                    put(rc,2.) !!
                                    '))');
               end;
               call symput('itms_rc','8');
               stop;
            end;

            rc = metadata_getattr(dirURI, "DirectoryName",
                 dir32);

            /* Verify if the directory location exists. */
            /* This is only used for the report.        */
            if (fileexist(dir32)) then
               alreadyExists = 1;

            output _pkglib2;

         end;

         else
         do;
           _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                              'sashelp.itms, ' !!
                              'RMDMUPKG_lib_not_found_1_err, ' !!
                              'noquote ' !!
                              '))');
            _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                              'sashelp.itms, ' !!
                              'RMDMUPKG_lib_not_found_2_err, ' !!
                              'noquote,' !!
                              strip(name) !!
                              ',' !!
                              strip(libid) !!
                              ',' !!
                              strip(libref) !!
                              '))');
            _msgtxt = resolve('%rmLogErr(ITMS,%rmmsg(' !!
                              'sashelp.itms, ' !!
                              'RMDMUPKG_lib_not_found_3_err, ' !!
                              'noquote ' !!
                              '))');
            call symput('itms_rc','8');
         end;

      end;  /* if found = 1 */

   run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;

  /* ------------------------- Unpackage ----------------------------*/

   %if &debug %then
   %do;
      title 'Work._pkglib2 Contents';
      proc print data = work._pkglib2;
      run;
   %end;

   %if (&doPackages = YES or &doPackages = INDEXONLY) %then
   %do;
      %unpack;
   %end;


   /* -------------------------- Write Report ------------------------*/

   %if (&doReport = YES) %then %do;

      %if %sysfunc(exist(work._pkgspeca_)) %then
      %do;
         proc sort data=_pkgspeca_ nodupkey equals;
            by dir;
         run;
      %end;

      %if %sysfunc(exist(work._pkgspecs_)) %then
      %do;
         proc sort data=_pkgspecs_ nodupkey equals;
            by dir;
         run;
      %end;

      data _null_;

         retain underline;

         %let rpt_title = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_title);
         %let rpt_subtitle = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_subtitle);
         %let rpt_divider_1 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_divider_1);
         %let rpt_divider_2 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_divider_2);
         %let rpt_nolibs_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_nolibs_msg);
         %let rpt_lib_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_label);
         %let rpt_libref_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_libref_label);
         %let rpt_metaid_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_metaid_label);
         %let rpt_outfile_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_outfile_label);
         %let rpt_old_lib_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_old_lib_label);
         %let rpt_new_lib_label = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_new_lib_label);
         %let asterisk = %rmmsg(sashelp.itms,
                                 RMDMUPKG_asterisk);
         %let rpt_new_lib_msg_1 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_new_lib_msg_1);
         %let rpt_new_lib_msg_2 = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_new_lib_msg_2);
         %let rpt_allocerr_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_allocerr_msg);
         %let rpt_lib_notfound_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_notfound_msg);
         %let rpt_lib_1_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_1_msg);
         %let rpt_lib_2_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_2_msg);
         %let rpt_lib_3_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_3_msg);
         %let rpt_lib_4_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_4_msg);
         %let rpt_lib_5_msg = %rmmsg(sashelp.itms,
                                 RMDMUPKG_rpt_lib_5_msg);

         file print;
         title "&rpt_title";

         set _pkglib2 nobs=nobs end=last;

         if (_n_ eq 1) then
         do;

            put / @5 "&rpt_subtitle";
            length underline $89;
            underline = repeat("&rpt_divider_1", 88);
            put @5 underline /;
            underline = repeat("&rpt_divider_2", 88);

            if nobs = 0 then do;
               put @5 "&rpt_nolibs_msg";
               stop;
            end;

         end;

         put @5 "&rpt_lib_label " name;
         if matched = 1 then
         do;

            put @5 "&rpt_libref_label " libref @23 "&rpt_metaid_label " libid;
            put @5 "&rpt_outfile_label " file32;
            put @5 "&rpt_old_lib_label " dir311;

            select (upcase(libref));

               when ("ADMIN")
               do;
                  %if %sysfunc(exist(work._pkgspeca_)) %then
                  %do;
                     do obsnuma = 1 to nobsa;
                        set _pkgspeca_ point=obsnuma nobs=nobsa;
                        if exists = 0 then
                           put @5 "&rpt_new_lib_label " dir " &asterisk";
                        else
                           put @5 "&rpt_new_lib_label " dir;
                     end;
                  %end;
                  %else
                  %do;
                     %if (&doPackages = YES or &doPackages = INDEXONLY) %then
                     %do;
                        put @5 "&rpt_new_lib_msg_1";
                     %end;
                     %else
                     %do;
                        put @5 "&rpt_new_lib_msg_2";
                     %end;
                  %end;
               end;

               when ("SPIN")
               do;
                  %if %sysfunc(exist(work._pkgspecs_)) %then
                  %do;
                     do obsnums = 1 to nobss;
                        set _pkgspecs_ point=obsnums nobs=nobss;
                        if exists = 0 then
                           put @5 "&rpt_new_lib_label " dir " &asterisk";
                        else
                           put @5 "&rpt_new_lib_label " dir;
                     end;
                  %end;
                  %else
                  %do;
                     %if (&doPackages = YES or &doPackages = INDEXONLY) %then
                     %do;
                        put @5 "&rpt_new_lib_msg_1";
                     %end;
                     %else
                     %do;
                        put @5 "&rpt_new_lib_msg_2";
                     %end;
                  %end;
               end;

               otherwise
               do;
                  if alreadyExists = 0 then do;
                     %if (&doPackages = YES or &doPackages = INDEXONLY) %then
                     %do;
                        if (fileexist(dir32)) then
                           put @5 "&rpt_new_lib_label " dir32 " &asterisk";
                        else
                           put @5 "&rpt_new_lib_label " dir32 " &rpt_allocerr_msg";
                     %end;
                     %else
                     %do;
                        put @5 "&rpt_new_lib_label " dir32 " &asterisk";
                     %end;
                  end;
                  else
                     put @5 "&rpt_new_lib_label " dir32;
               end;

            end;  /* end select */

         end;  /* if matched */

         else
         do;
            put @5 "&rpt_lib_notfound_msg";
         end;

         put @5 underline;

         %if (&doPackages = YES or &doPackages = INDEXONLY) %then
         %do;
            if last then
            do;
               put / @5 nobs " &rpt_lib_1_msg" /;
               put @5 "&rpt_lib_2_msg";
            end;
         %end;
         %else
         %do;
            if last then
            do;
               put / @5 nobs " &rpt_lib_3_msg";
               put @5 "&rpt_lib_4_msg" /;
               put @5 "&rpt_lib_5_msg";
            end;
         %end;

      run;

   %end;


   /* ----------------------------- Exit -----------------------------*/
   %exit:

   %if &itms_rc = 4 %then %do;
      %rmLogWrn(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_proc_warn,
                             noquote,
                             &itms_rc
                             ));
      %rmLogWrn(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_review_warn,
                             noquote
                             ));
   %end;

   %if &itms_rc > 4 %then %do;
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_proc_err,
                             noquote,
                             &itms_rc
                             ));
      %rmLogErr(ITMS, %rmmsg(sashelp.itms,
                             RMDMUPKG_review_err,
                             noquote
                             ));
   %end;

   %if &debug ne 1 %then
   %do;
      proc datasets lib=work nolist nowarn;
         delete _pkg:;
      run;
      quit;
   %end;

   /* set return code in user-specified macro variable */
   %if &_rc ne %str() %then %do;
      %let &_rc = &itms_rc;
   %end;

   /* set rc for batch */
   %if &itms_rc gt 0 and &sysenv = BACK %then
   %do;
      options noerrorabend;
      data _null_;
         abort &itms_rc;
      run;
   %end;

%mend rmdmupkg;
