/**********************************************************************/
/* Macro: rmdmpkg                                                     */
/*        IT Data Mart Package macro                                  */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Description:                                                       */
/* This macro is normally run on an earlier release of ITRM, for      */
/* example, ITRM 3.1.1 or 3.2, to package all the physical SAS data   */
/* libraries that compose a single IT datamart.  The package is       */
/* transferred to another system and the %rmdmupkg macro is           */
/* subsequently run against that package using a later release of     */
/* ITRM, for example, ITRM 3.3, to un-package the libraries. This     */
/* is the typical usage during migration, however, it may also be     */
/* used to move data between two ITRM releases that are at the same   */
/* release level.                                                     */
/*                                                                    */
/* The packages are in transport format and are created with          */
/* PROC CPORT.  The macro may be run in report mode first so that     */
/* only a report of what would be done is produced.  This is          */
/* the default.  To actually create the packages in the output        */
/* directory, specify doPackages=YES.                                 */
/*                                                                    */
/* The rmdmpkg macro is typically followed by an execution of the     */
/* rmdmpath and rmdmupkg macros. This macro should be executed on     */
/* the ITRM server tier.                                              */
/*                                                                    */
/* The macro supports packaging SAS datasets and catalogs, but        */
/* not SAS views.                                                     */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Parameters:                                                        */
/*     metaserver - Metadata server host name                         */
/*          -REQUIRED-                                                */
/*     metaport - Metadata server port number (usually 8561)          */
/*          -REQUIRED-                                                */
/*     metauser - Userid to access the metadata server                */
/*          -REQUIRED-                                                */
/*     metapass - Password for the metauser userid                    */
/*          -REQUIRED-                                                */
/*     datamart - Name of the input IT datamart that is               */
/*               to be used.                                          */
/*          -REQUIRED-                                                */
/*     pkgdir -  Output directory to contain the CPORT files.         */
/*               This directory must exist prior to running           */
/*               %rmdmpkg and must have enough disk space available   */
/*               to accommodate all IT data mart physical datasets    */
/*               in transport format. For traditional z/OS, this      */
/*               should be a PDS or PDS/E.                            */
/*          -REQUIRED-                                                */
/*     idxname - name of the file that is to be used as an            */
/*               index of the files that are packaged.  This          */
/*               parameter is optional and only needs to be supplied  */
/*               if there are naming conflicts.  The default name     */
/*               is PKGIDX.cpt on Windows and Unix and PKGIDX on z/OS.*/
/*     repos - Name of the repository containing the IT data mart.    */
/*               This parameter is optional, the default is           */
/*               FOUNDATION.                                          */
/*     stagedLibs - Whether or not to migrate staged table libraries. */
/*               This parameter is optional, valid values are YES or  */
/*               NO, the default is NO.                               */
/*     doReport - Create a report detailing what data libraries       */
/*               are to be packaged.  This parameter is optional,     */
/*               valid values are YES or NO, the default is YES.      */
/*     doPackages - Create the CPORT packages from the SAS data       */
/*               libraries.  This parameter is optional,              */
/*               valid values are YES or NO, the default is NO.       */
/*     disp - status of the input datasets for z/OS only.  This       */
/*               parameter is optional and defaults to OLD.  Valid    */
/*               values are OLD and SHR.                              */
/*     _rc - user-specified return code macro variable.  This         */
/*               parameter is optional.                               */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Error Recovery                                                     */
/* To rerun the macro in the event of an error, clear out the output  */
/* directory pointed to by the pkgdir= parameter and rerun.  This     */
/* macro does not update any metadata on the SAS Metadata Server.     */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* History                                                            */
/* 12Jan2012 rmdmpkg version 303000                                   */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Copyright (c) 2009, 2010 SAS Institute, Inc Cary, NC USA           */
/* All rights reserved                                                */
/*                                                                    */
/**********************************************************************/

%macro rmdmpkg (metaServer=,
                metaPort=8561,
                metaUser=,
                metaPass=,
                datamart=,
                pkgdir=,
                idxname=PKGIDX,
                repos=FOUNDATION,
                stagedLibs=NO,
                doReport=YES,
                doPackages=NO,
                disp=OLD,
                _rc=);

   /* TODO: message stuff S0640530
   %global RMLOGGERLEVEL;
   %let RMLOGGERLEVEL=2;
   %rmcrmsg;
   -----*/

   %global debug;
   %local itms_rc;
   %local _pkgscp;
   %local _pkgdelim;
   %local _idxpath;
   %local _libdisp;
   %local _version;

   %let _version = 303000;
   %let itms_rc = 0;

   %if (&debug ne 1) %then
      %let debug = 0;

   %if ("&sysscp" eq "WIN") %then
   %do;
      %let _pkgscp = WIN;
      %let _pkgdelim = %str(\);
   %end;

   %else
   %do;
      %if ("&sysscp" eq "OS") %then
      %do;
         %if %index(%quote(&pkgdir),%str(/)) ne 0 %then
         %do;
            %let _pkgscp = ZFS;
            %let _pkgdelim = %str(/);
         %end;
         %else
         %do;
            %let _pkgscp = OS;
            %let _pkgdelim = %str(.);
         %end;
      %end;

      %else
      %do;
         %let _pkgscp = UNX;
         %let _pkgdelim = %str(/);
      %end;
   %end;

   %if &debug %then
       %put SYSSCP is &sysscp and SCP is &_pkgscp;

   %if &sysver ne 9.1 and &sysver ne 9.2 and &sysver ne 9.3 and &sysver ne 9.4 %then
   %do;
      %let _t1 = %str(macro must be executed);
      %let _t2 = %str(on a SAS 9.1 system or higher);
      %put ERROR%quote(:) The %nrquote(%quote(%%)rmdmupkg) &_t1 &_t2;
      %let itms_rc=8;
      %goto exit;
   %end;


   /*--------------------- Verify Parameters  -----------------------*/

   %if (%quote(&metaServer) eq ) %then
   %do;
      %put ERROR%quote(:) The metaServer parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (&metaPort eq ) %then
   %do;
      %put ERROR%quote(:) The metaPort parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&metaUser) eq ) %then
   %do;
      %put ERROR%quote(:) The metaUser parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&metaPass) eq ) %then
   %do;
      %put ERROR%quote(:) The metaPass parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&datamart) eq ) %then
   %do;
      %put ERROR%quote(:) The datamart parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;
   %if (%quote(&pkgdir) eq ) %then
   %do;
      %put ERROR%quote(:) The pkgdir parameter is required;
      %let itms_rc=8;
      %goto exit;
   %end;

   %if (%quote(&idxname) eq ) %then
      %let idxname = PKGIDX;
   %if &_pkgscp = OS %then
      %let _idxpath = %qtrim(&pkgdir.)%str(%()&idxname.%str(%));
   %else
      %let _idxpath = %qtrim(&pkgdir.)&_pkgdelim.&idxname..cpt;

   %if (%quote(&repos) eq ) %then
      %let repos = FOUNDATION;

   %if (&stagedLibs eq ) %then
      %let stagedLibs = NO;
   %else
      %let stagedLibs = %upcase(&stagedLibs);
   %if (&stagedLibs ne NO and &stagedLibs ne YES) %then
   %do;
      %put ERROR%quote(:) The stagedLibs parameter must be YES or NO;
      %let itms_rc=8;
      %goto exit;
   %end;

   %if (&doReport eq ) %then
      %let doReport = YES;
   %else
      %let doReport = %upcase(&doReport);
   %if (&doReport ne NO and &doReport ne YES) %then
   %do;
      %put ERROR%quote(:) The doReport parameter must be YES or NO;
      %let itms_rc=8;
      %goto exit;
   %end;

   %if (&doPackages eq ) %then
      %let doPackages = NO;
   %else
      %let doPackages = %upcase(&doPackages);
   %if (&doPackages ne NO and &doPackages ne YES and &doPackages ne INDEXONLY) %then
   %do;
      %put ERROR%quote(:) The doPackages parameter must be YES or NO;
      %let itms_rc=8;
      %goto exit;
   %end;


   %if (&disp eq ) %then
      %let disp = OLD;
   %else
      %let disp = %upcase(&disp);
   %if (&disp ne SHR and &disp ne OLD) %then
   %do;
      %put ERROR%quote(:) The disp parameter must be SHR or OLD;
      %let itms_rc=8;
      %goto exit;
   %end;

   %if ("&sysscp" eq "OS") %then
   %do;
      %let _libdisp = %str(DISP=&disp);
   %end;

   %else
   %do;
      %let _libdisp = ;
   %end;

   /*--------------------- Macro Definitions  -----------------------*/

   /* Encode strings to current locale for z/OS */
   %macro encodeValue(stringVar);
      %if ("&_pkgscp" eq "OS") %then
      %do;
         length encoding $20;
         encoding = getoption("ENCODING");
         if (encoding not in ("EBCDIC1047", "OPEN_ED-1047")) then
            &stringVar = kcvt(&stringVar, "EBCDIC1047", encoding);
      %end;
   %mend encodeValue;


   /* DATA step fragment to set the libType variable in the         */
   /* _pkglib_ dataset.                                             */
   /* libType = 0  undetermined type, e.g., ADMIN or SPIN           */
   /* libType = 1  aggregation library, either summary or detail    */
   /* libType = 2  staged library to be migrated                    */
   /* libType = 3  staged library not to be migrated                */
   %macro setLibType(libURI);

         length tblrole $40;
         length roleURI tblURI $256;
         length libType 8;

         tblrole = "";
         roleURI = "";
         tblURI = "";
         libType = 0;

         tblrc = 1;
         tblcount = 1;
         tblrc = metadata_getnasn(libURI, "Tables",
                                  tblcount, tblURI);
         if (tblrc > 0) and (tblcount ge 1) then
         do;
            libType = 0;
            do i = 1 to tblcount;
               rc = metadata_getnasn(tblURI, "Roles",
                                     1, roleURI);
               if roleURI ne "" then
               do;
                  rc = metadata_getattr(roleURI, "ObjectRole",
                                        tblrole);

                  if upcase(tblrole) =: "SUMMARY" or upcase(tblrole) =: "DETAIL" then
                     libType = 1;
                  if upcase(tblrole) =: "STAGE" and symget('stagedLibs') = "YES" then
                     libType = 2;
                  if upcase(tblrole) =: "STAGE" and symget('stagedLibs') = "NO" then
                     libType = 3;

               end;
            end;
         end;

   %mend setLibType;


   /* This DATA step fragment is called by %pack macro to retrieve  */
   /* aging column info.  It sets the agingcol and agingfmt vars.   */
   %macro agingcol(tblURI);

      colrc = 1;
      colcount = 1;
      do while (colrc > 0);
         colrc = metadata_getnasn(&tblURI, "Columns", colcount, colURI);
         if colrc > 0 then
         do;

            keyrc = metadata_getnasn(colURI, "Keywords", 1, keyURI);
            if keyrc > 0 then
            do;
               rc = metadata_getattr(keyURI, "Name", keyname);
               if keyname = "AColumn" then
               do;
                  rc = metadata_getattr(colURI, "Name", agingcol);
                  rc = metadata_getattr(colURI, "SASFormat", agingfmt);
                  rc = metadata_getattr(colURI, "Desc", aginglbl);
                  colrc = -99;
               end;
            end;

         end;

         colcount + 1;

      end;

   %mend agingcol;


   /* This macro is used to package all libraries but it also */
   /* retrieves table information for each table in the       */
   /* libraries that are packaged.  The table information     */
   /* is written to the _pkgtbl_ dataset for eventual use     */
   /* by the %rmdmupkg macro.                                 */
   %macro pack();

      /* clear work before starting */
      proc datasets lib=work nolist nowarn;
         delete _pkgtbl_;
         delete _pkgdic_;
         delete _pkgtbl_temp;
      run;
      quit;

      /* Process all libraries in _pkglib_ into packages */
      %let dsid = %sysfunc(open(work._pkglib_,i));
      %let nobs = %sysfunc(attrn(&dsid,NOBS));
      %if &nobs = 0 %then
      %do;
         %let _text1 = %str(There are no libraries in the );
         %let _text2 = %str(IT data mart &datamart. to package);
         %put WARNING: &_text1 &_text2;
         %let itms_rc = 4;
         %goto exit_pack;
      %end;

      /* this statement sets many of the macro vars used below */
      %syscall set(dsid);

      %let eof = 0;
      %do %until(&eof ne 0);

         %let rc = %sysfunc(fetch(&dsid));

         %if &rc = -1 %then
            %let eof = 1;

         %else %do;

            %if &found = 1 %then
            %do;

               libname &libref "%trim(&dir311)" &_libdisp;

               /* Get the table names in the library */
               proc sql;
                  create table _pkgdic_ as
                  select memname
                  from dictionary.tables
                  where trim(upcase(libname)) ="%trim(%upcase(&libref))";
               quit;

               /* Create table information for each table    */
               /* encountered in the library.                */
               data _pkgtbl_temp (keep = libref libid
                                         tblname tblid tblrole
                                         aggtbl agingcol agingfmt
                                         aginglbl);

                  length libURI tblURI roleURI colURI keyURI $256;
                  length aginglbl $256;
                  length tblrole keyname tblname_meta uri $40;
                  length tblname aggtbl agingcol agingfmt $32;
                  length tblid libid id $17;
                  length libref $8;

                  libURI = "";
                  tblURI = "";
                  roleURI = "";
                  colURI = "";
                  keyURI = "";
                  tblrole = "";
                  keyname = "";
                  aggtbl = "";
                  tblid = "";
                  libid = "";
                  agingcol = "";
                  aginglbl = "";
                  agingfmt = "";
                  tblname_meta = "";

                  set _pkgdic_;

                  libref = "&libref";
                  libid = "&libid";
                  tblname = memname;

                  /* Info that is specific to 3.1.1 summary status   */
                  /* tables.  These tables will require reformatting */
                  /* by rmdmupkg, so are written to _pkgtbl_ even    */
                  /* though they are not registered in metadata.     */
                  if tblname =: "STATUS_" then
                  do;
                     tblrole = "SummaryStatus";

                     /* get the corresponding agg table */
                     id = substr(tblname,8,17);
                     id = translate(id,".","_");
                     uri = "OMSOBJ:PhysicalTable\" !! id;
                     tblrc = metadata_getnobj(uri, 1, tblURI);

                     /* Get some aging column info */
                     if tblrc > 0 then do;
                        rc = metadata_getattr(tblURI, "SASTableName", aggtbl);
                        %agingcol(tblURI);
                     end;

                     output _pkgtbl_temp;
                  end;

                  /* Staged, summary, and other tables fall thru to here. */
                  /* Tables that are resident in the library but are not  */
                  /* registered to metadata will not get written to       */
                  /* _pkgtbl_, however, will get cported to the target    */
                  /* system when unpacked by rmdmupkg.                    */
                  else
                  do;

                     /* get the SASLibrary object for this library */
                     rc = metadata_getnobj("OMSOBJ:SASLibrary\&libid",
                          1, libURI);

                     /* loop thru the metadata retrieving table info */
                     tblrc = 1;
                     tblcount = 1;
                     do while (tblrc > 0);

                        tblrc = metadata_getnasn(libURI, "Tables",
                                                 tblcount, tblURI);
                        if tblrc > 0 then
                        do;
                           rc = metadata_getattr(tblURI, "SASTableName",
                                                 tblname_meta);

                           if upcase(tblname_meta) = upcase(tblname) then
                           do;
                              rc = metadata_getattr(tblURI, "Id", tblid);
                              rc = metadata_getnasn(tblURI, "Roles",
                                                    1, roleURI);
                              rc = metadata_getattr(roleURI, "ObjectRole",
                                                    tblrole);

                              /* Retrieve aging column info if summary */
                              if upcase(tblrole) =: "SUMMARY" or upcase(tblrole) =: "DETAIL" then
                              do;
                                 %agingcol(tblURI);
                              end;

                              output _pkgtbl_temp;

                           end;

                        end;

                        tblcount + 1;

                     end;

                  end;

               run;

               /* append the table info on to the main table dataset */
               proc datasets lib=work nolist;
                  append base=_pkgtbl_ data=_pkgtbl_temp;
                  delete _pkgdic_;
                  delete _pkgtbl_temp;
               run;
               quit;

               /* Packages are not created for INDEXONLY runs.    */
               /* Packages for staging libraries are not created  */
               /* if stagedLibs=NO (libType = 3).                 */
               %if (&doPackages = YES and &libType ne 3) %then %do;

                  /* cport the entire library */
                  filename _pkgout "%trim(&file311)";

                  %if %sysfunc(exist(&libref..sasmacr,CATALOG)) %then
                  %do;
                     proc cport lib = &libref
                                file = _pkgout;
                          exclude sasmacr/mt=catalog;
                     run;
                  %end;

                  %else
                  %do;
                     proc cport lib = &libref
                                file = _pkgout;
                     run;
                  %end;

                  filename _pkgout clear;

               %end;

               libname &libref clear;

            %end;  /* if found = 1 */

         %end;

      %end;

      %let rc = %sysfunc(close(&dsid));

   %exit_pack:
   %mend pack;

   /*----------------- End of Macro Definitions  --------------------*/


   /*------------------------  Setup --------------------------------*/


   options metaserver="&metaserver"
           metaport=&metaport
           metauser="&metauser"
           metapass="&metapass"
           metaRepository="&repos";


   /*------------------- Validate Output Directory ------------------*/

   data _null_;

      if (not(fileexist("&pkgdir"))) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         put "ERROR: The external file &pkgdir does not exist.";
         put "ERROR: The directory pointed to by the pkgdir= "
             "parameter must exist prior to calling this macro.";
         put "ERROR: Verify that the value for the pkgdir= parameter "
             "is spelled correctly";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;

      rc = filename("_pkgout", "&pkgdir");
      if (rc) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         put "ERROR: The external file &pkgdir could not be "
             "allocated.";
         put "ERROR: Verify that the directory pointed to by the "
             "pkgdir= parameter is available.";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;

      /* Prove that its a directory */
      dirid = dopen('_pkgout');
      if (not dirid) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         put "ERROR: The external file &pkgdir is not a directory.";
         put "ERROR: Verify that the value of the pkgdir= parameter "
             "is a directory.";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;

      rc = dclose(dirid);
      if (rc) then
      do;
         _msg_ = sysmsg();
         put _msg_;
         put "ERROR: An error was returned when closing the "
             "directory &pkgdir.";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;

      /* Prove that we can write a file to the output directory       */
      /* before proceeding, but don't do this if this is a            */
      /* report-only run.                                             */
      %if (&doPackages = YES or &doPackages = INDEXONLY) %then %do;
         rc = filename("_pkgout", "");
         rc = filename("_pkgout", "&_idxpath");
         if (rc) then
         do;
            _msg_ = sysmsg();
            put _msg_;
            put "ERROR: Unable to create a filename reference to the "
                "index file in &pkgdir.";
            put "ERROR: Ensure that authorization to this directory "
                "is sufficient.";
            _error_ = 0;
            call symput('itms_rc','8');
            stop;
         end;

         rc = fileref ("_pkgout");
         if rc > 0 then
         do;
            _msg_ = sysmsg();
            put _msg_;
            put "ERROR: Unable to create a fileref reference to the "
                "index file in &pkgdir.";
            put "ERROR: Ensure that authorization to this directory "
                "is sufficient.";
            _error_ = 0;
            call symput('itms_rc','8');
            stop;
         end;

         else
         do;
            if rc = 0 then
               put "NOTE: Existing package file will be "
                   "overwritten.";
            fid = fopen("_pkgout", "o");
            if fid = 0 then
            do;
               _msg_ = sysmsg();
               put _msg_;
               put "ERROR: Unable to open and write the index file "
                   "to &pkgdir.";
               put "ERROR: Ensure that authorization to this directory "
                   "is sufficient.";
               _error_ = 0;
               call symput('itms_rc','8');
               stop;
            end;
            rc = fclose(fid);
            if rc ne 0 then
            do;
               _msg_ = sysmsg();
               put _msg_;
               put "ERROR: Unable to close and write the index file "
                   "to &pkgdir.";
               put "ERROR: Ensure that authorization to this directory "
                   "is sufficient.";
               _error_ = 0;
               call symput('itms_rc','8');
               stop;
            end;
         end;
      %end;

      rc = filename("_pkgout", "");

   run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;


   /*------------------- Validate IT Datamart -----------------------*/

   data _pkgdm_(keep=itdmURI);

      length searchString $256;
      length itdmURI $256;

      itdmURI = "";

      /* If 9.1, then look for a tree, if 9.2+ look for a group */
      %if &sysver = 9.1 %then
      %do;
         searchString = "omsobj:Tree?" !!
                        "Tree[@Name eq '&datamart']" !!
                        "[Extensions/Extension[@Value eq 'ITMS:DataMart']]";
         %encodeValue(searchString);
      %end;
      %else
      %do;
         searchString = "omsobj:Group?" !!
                        "Group[@Name eq '&datamart' and " !!
                        "@PublicType eq 'ITDataMart']";
      %end;

      /* get the start of the data mart tree */
      nobj = metadata_getnobj(searchString, 1, itdmURI);

      if (nobj eq -1) then
      do;
         put "ERROR: Could not establish connection to metadata "
             "server, unable to continue.";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;

      if (itdmURI eq "") then
      do;
         put "ERROR: Unable to find IT data mart, &datamart, "
             "in the metadata.";
         put "ERROR: Verify that the IT data mart name is spelled "
             "correctly, case is correct, and";
         put "ERROR: the metadata parameters are correct.";
         _error_ = 0;
         call symput('itms_rc','8');
         stop;
      end;
      else
      do;
         output _pkgdm_;
      end;

      %if &debug %then
      %do;
         put 'After getnobj of datamart ' itdmURI= nobj=;
      %end;

    run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;


   /*--------------- Get List of Libraries to Package ---------------*/

   data _pkglib_(keep=name libid libref dir311 file311 filename
                      found libType ver_src);

      length parentURI treeURI libURI dirURI $256;
      length name $60;
      length libid $17;
      length libref $8;
      length dir311 $256;
      length file311 $256;
      length filename $20;
      length ver_src $6;
      length found 8;

      parentURI = "";
      treeURI = "";
      libURI = "";
      dirURI = "";
      name = "";
      libid = "";
      libref = "";
      dir311 = "";
      file311 = "";
      filename = "";
      found = 0;
      ver_src = "&_version";

      /* arrays to handle the recursion */
      array parents [1000] $ 256 _temporary_;
      array trees [1000] $ 256 _temporary_;
      array treeCount [1000] _temporary_;

      set _pkgdm_;
      parentURI = itdmURI;

      %if &debug %then
      %do;
         put itdmURI=;
      %end;

      /* Must get the top of tree first if running under 9.2+ */
      %if &sysver = 9.1 %then
      %do;
         parentURI = itdmURI;
      %end;

      %else %do;
         rc = metadata_getnasn(itdmURI, "Trees",
                               1, parentURI);
         %if &debug %then
         %do;
            put 'After getnasn of itdmURI for 9.2+ ' parentURI= rc=;
         %end;
      %end;


      /* start recursive routine */
      level = 0;
      link getLibs;

      return;

      /* recursive section - traverse trees and look for libraries */
      getLibs:

         /* get the subtrees from a given parent tree */
         tree = 0;
         do until (treeURI eq "");

            tree + 1;
            treeURI  = "";

            %if &debug %then
            %do;
               put 'Before getnasn of subtrees ' parentURI=;
            %end;

            rc = metadata_getnasn(parentURI, "SubTrees",
                                  tree, treeURI);

            %if &debug %then
            %do;
               put 'After getnasn of subtrees ' treeURI= rc=;
            %end;

            if (treeURI ne "") then
            do;

               /* get members of the tree and look for libraries */
               mem = 0;
               do until (libURI = "");

                  mem + 1;
                  libURI = "";
                  nlib = metadata_getnasn(treeURI, "Members",
                                        mem, libURI);

                  %if &debug %then
                  %do;
                     put 'After getnasn of tree members ' libURI= nlib=;
                  %end;

                  if (libURI ne "") then
                  do;

                     /* Is this tree member a library? */
                     if (scan(libURI, 2, ":\") eq "SASLibrary") then
                     do;

                        rc = metadata_getattr(libURI, "Name", name);
                        rc = metadata_getattr(libURI, "ID", libid);
                        rc = metadata_getattr(libURI, "Libref", libref);

                        /* set the libType variable */
                        %setLibType(libURI);
                        %if &debug %then
                        %do;
                           put 'After setLibType macro ' libType=;
                        %end;

                        rc = metadata_getnasn(libURI, 'UsingPackages',
                             1, dirURI);

                        if (rc le 0) then
                        do;
                           if rc = -1 then do;
                              put "ERROR: Lost connection to metadata "
                                  "server, unable to continue.";
                              _error_ = 0;
                              call symput('itms_rc','8');
                              stop;
                           end;
                           else do;
                              put "WARNING: Unable to retrieve "
                                  "required metadata for Library, ";
                              put "WARNING:" rc= name= libid= libref= ;
                              call symput('itms_rc','4');
                           end;
                        end;

                        else do;

                           /* Is this a directory? */
                           if (scan(dirURI, 2, ":\") eq "Directory") then
                           do;

                              rc = metadata_getattr(dirURI,
                                            "DirectoryName", dir311);
                              if (dir311 ne "") then
                              do;
                                 if libType ne 3 then
                                 do;
                                    if "&_pkgscp" = "OS" then
                                    do;
                                       filename = strip(libref);
                                       file311 = strip("&pkgdir") !! "(" !!
                                                 strip(filename) !! ")";
                                    end;
                                    else
                                    do;
                                       filename = strip(libref) !! ".cpt";
                                       file311 = strip("&pkgdir") !! "&_pkgdelim" !!
                                                 strip(filename);
                                    end;
                                 end;
                                 else
                                 do;
                                    filename = "";
                                    file311 = "***transport file not created " !!
                                              "due to stagedLibs=NO option";
                                 end;

                                 /* try to allocate the directory */
                                 librc = libname(libref, dir311,,"&_libdisp");
                                 if (librc = 0) then
                                 do;
                                    found = 1;
                                    librc = libname(libref);
                                 end;
                                 else
                                    found = 0;

                                 output _pkglib_;
                              end;

                              else
                              do;
                                 put "WARNING: Unable to retrieve "
                                     "directory for Library, ";
                                 put "WARNING:" rc= name= libid= libref= ;
                                 call symput('itms_rc','4');
                              end;

                           end;  /* if directory */

                        end;  /* rc ge 0 */

                     end;  /* if saslibrary */

                  end;  /* if - non blank member uri */

               end;  /* do until - no more members */

               /* save current level information in arrays */
               /* before recursing                         */
               level + 1;
               parents[level] = parentURI;
               trees[level] = treeURI;
               treeCount[level] = tree;
               parentURI = treeURI;

               link getLibs;

               /* restore current level information from */
               /* arrays after recursing                 */
               treeURI = trees[level];
               parentURI = parents[level];
               tree = treeCount[level];
               level = level - 1;

            end;  /* if - non blank sub tree uri */

         end;  /* do until - no more subtrees */

      return;

   run;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;

   %if &debug %then
   %do;
      title 'Work._pkglib_ Contents before pack macro';
      proc print data = work._pkglib_;
      run;
   %end;


   /*---------------------- Package the Data -------------------------*/

   %if (&doPackages = YES or &doPackages = INDEXONLY) %then %do;
      %pack;
   %end;

   %if &itms_rc > 4 %then
   %do;
      %goto exit;
   %end;


   /*------------------ Package the Index File -----------------------*/

   %if &debug %then
   %do;
      title 'Work._pkglib_ Contents';
      proc print data = work._pkglib_;
      run;

      %if (&doPackages = YES or &doPackages = INDEXONLY) %then
      %do;
         title 'Work._pkgtbl_ Contents';
         proc print data = work._pkgtbl_;
         run;
      %end;
   %end;

   %if (&doPackages = YES or &doPackages = INDEXONLY) %then %do;

      /* The %pack macro executed above provided all the info    */
      /* for the tables in each library to the _pkgtbl_ dataset. */

      filename _pkgout "&_idxpath";

      proc cport lib=work file=_pkgout;
         select _pkglib_;
         select _pkgtbl_;
      run;

      filename _pkgout clear;

   %end;


   /*-------------------------- Write Report ------------------------*/

   %if (&doReport = YES) %then %do;

      data _null_;

         retain underline;
         retain f 0 nf 0;

         file print;
         title "IT Data Mart Package Report";

         %if (&doPackages = YES) %then
         %do;
            %let _text1 = %str(ITRM libraries packaged);
            %let _text2 = %str(from IT data mart &datamart);
            %let _text = &_text1 &_text2;
            title2 "&_text";
         %end;
         %else
         %do;
            %let _text1 = %str(ITRM libraries that could be);
            %let _text2 = %str(packaged from IT data mart &datamart);
            %let _text = &_text1 &_text2;
            title2 "&_text";
         %end;

         set _pkglib_ nobs=nobs end=last;

         if (_n_ eq 1) then
         do;
            underline = repeat("-", 88);
            put /;
            if nobs = 0 then do;
               put @5 "There are no libraries in the "
                      "IT data mart &datamart. to package";
               stop;
            end;
         end;

         put @5 "Library Name: " name;
         put @5 "libref: " libref @23 "Metadata ID: " libid;
         if found = 1 then
         do;
            put @5 "Library Location: " dir311;
            if libType ne 3 then
            do;
               put @5 "Output Transport File Name: " file311;
               f = f + 1;
            end;
            else
            do;
               put @5 "Output Transport File Name: "
                      "***library will not be packaged due to "
                      "stagedLibs=NO option";
            end;
         end;
         else
         do;
            put @5 "Library Location: " dir311
                   " ***not found";
            put @5 "Output Transport File Name: " file311
                   " ***not built";
         end;
         put @5 underline;

         %if (&doPackages = YES) %then
         %do;
            if last then do;
               put / @5 f " total libraries were packaged";
               if f ne nobs then do;
                  nf = nobs - f;
                  put @5 nf
                      " libraries were not packaged, see messages above";
               end;
            end;
         %end;
         %else
         %do;
            if last then
               put / @5 nobs " total libraries were encountered; "
                          "specify doPackage=YES to create the "
                           "transport packages.";
         %end;

      run;

   %end;


   /*----------------------------- Exit -----------------------------*/
   %exit:

   %if &itms_rc = 4 %then %do;
      %let _text = %str(One or more processing warnings have occurred);
      %put WARNING%quote(:) &_text;
      %put WARNING%quote(:) itms_rc = &itms_rc;
      %put WARNING%quote(:) Please review the SAS log for warnings;
   %end;

   %if &itms_rc > 4 %then %do;
      %let _text = %str(One or more processing errors have occurred);
      %put ERROR%quote(:) &_text;
      %put ERROR%quote(:) itms_rc = &itms_rc;
      %put ERROR%quote(:) Please review the SAS log for errors;
   %end;

   %if &debug ne 1 %then
   %do;
      proc datasets lib=work nolist nowarn;
         delete _pkg:;
      run;
      quit;
   %end;

   /* set return code in user-specified macro variable */
   %if &_rc ne %str() %then %do;
      %let &_rc = &itms_rc;
   %end;

   /* set rc for batch */
   %if &itms_rc gt 0 and &sysenv = BACK %then
   %do;
      options noerrorabend;
      data _null_;
         abort &itms_rc;
      run;
   %end;

%mend rmdmpkg;
