%macro RMDUPCHK(source=,
                idvar=,
                timestmp=,
                endfile=,
                int=00:29,
                report=YES,
                systems=,
                ranges=,
                keep=9,
                force=NO,
                term=NO);

 /*-------------------------------------------------------------------+
  | Copyright (C) 1995 by SAS Institute Inc., Cary, NC 27512-8000
  | Name:     RMDUPCHK.sas
  | Support:  sasccw
  | Product:  SAS IT Resource Management
  | Language: SAS
  | Purpose:  Check for duplicate incoming data and delete it. Also,
  |           build up record counts of incoming and deleted data and
  |           datetime ranges for each system or machine.
  |
  | Input:    Parm source      - The type of data to be screened. Must
  |                              be 3 characters.
  |           Parm idvar       - The name of a SAS variable which
  |                              identifies the system or machine
  |                              which generated the input data.
  |           Parm timestmp    - The name of the SAS variable which
  |                              contains the datetimestamp which
  |                              uniquely identifies the time at
  |                              which each record was produced.
  |           Parm endfile     - The name of the SAS variable that is
  |                              used as the END= keyword on the
  |                              INFILE statement which reads the raw
  |                              data.
  |           Parm int         - The maximium time gap (interval)
  |                              which the data duplication check
  |                              routines will tolerate. The default
  |                              is 00:29 (29 minutes).
  |           Parm systems     - The number of systems/machines from
  |                              which you expect to process, data both
  |                              in the new incoming data and what has
  |                              already been encountered and stored in
  |                              the control dataset.  The default is
  |                              three.
  |                              This parameter is no longer required
  |                              and will be ignored. It remains in the
  |                              macro definition for compatibility.
  |           Parm ranges      - The number of ranges of data you
  |                              expect to encounter, both in the new
  |                              incoming data and what has already
  |                              been encountered and stored in the
  |                              control dataset.  Gaps in the data
  |                              will be tolerated up to the time
  |                              specified by the INT parameter. Any
  |                              gap larger than this will imply the
  |                              start of a second range. These ranges
  |                              are stored in memory slots and need to
  |                              be pre-allocated. Gaps in the data
  |                              should be an exceptional event but you
  |                              should allow for the possibility to
  |                              prevent needless reruns due to
  |                              insufficient memory slots. The default
  |                              is 20.
  |                              This parameter is no longer required
  |                              and will be ignored. It remains in the
  |                              macro definition for compatibility.
  |           Parm keep        - The number of "dtweeks" for which
  |                              control data will be kept. Because
  |                              dtweeks counts the number of Sundays
  |                              between two dates, a value of 2 (the
  |                              default) actually gives you a
  |                              maximum retention period of 20 days.
  |           Parm force       - Controls whether detected duplicate
  |                              input data is allowed through. The
  |                              default is NO.
  |           Parm term        - Controls whether SAS terminates if
  |                              duplicate input data is detected.
  |                              The default is NO.
  |           Libname &DUPLIB  - The previous control data set is
  |                              concatenated with the new data.
  |
  | Output:   Report in log displaying system/machine names for which
  |           data has been read. This will show warnings and record
  |           counts if duplicate input data has been detected.  Macro
  |           variables containing the datetime ranges for each
  |           system/machine are also created.
  | Usage:
  | History:  Date    Action
  |           19Sep95 Initial coding                             CDW
  |           11Jan96 Replace global macro vars with a dataset
  |                   to pass info to RMDUPUPD                   CDW
  |           01Apr96 Changed references to CMPROCES DUPMODE
  |                   parameter to be correct  - D0098038        CDW
  |           22Jan97 Performance hit - use arrays, not formats,
  |                   to hold control data     - D0119053        CDW
  |           28Aug97 Globalise the CPCOLECT macro - D0129234    CDW
  |           09Sep97 Improve parm checking - D0134933           CDW
  |           01Jul98 D0122050 - only abort if in batch          CDW
  |           01Jul98 S0002126 - implement EOFONLY settings when
  |                   when used with RECFM=N, eg SAP             CDW
  |           01Jul98 S0004808 - ignore missing values for IDVAR
  |                   to avoid array subscript out of bounds msg CDW
  |           17Aug98 S0021893 Change STOP statements to ABORT
  |                   for interactive processing. This stops
  |                   processing immediately with no output ds.  TRN
  |           01Jul98 S0032346 - allow for datetime ranges to be
  |                   appended at the front. Also allow previous
  |                   ranges to be tested.                       CDW
  |           02Oct98 S0032949 Only abort abend for MVS due to
  |                   implications with JSWiz. Reset dsn and set
  |                   macros variable on failure so process can
  |                   rerun in interactive mode.                 TRN
  |           11Feb99 S0044040 Remove the use of nodsnferr and
  |                   use exist function instead. This allows us
  |                   to create the staging dataset as a view.   CDW
  |           19Feb99 S0044045 Add new CPABONXT macro variable
  |                   for collectors like SAP who need to abort
  |                   on the next datastep iteration.            CDW
  |           26Oct99 S0061754 - change NOTE to Note             CDW
  |           23Jan02 upgraded SYSTEMS= to support 99999 values  CDW
  |           07Apr03 change itsv to itrm                        CDW
  |           16Jun05 S0305206 Get FRCTEXT and TRMTEXT correct   CDW
  |           05Apr06 S0164856 upgraded RANGES= to support 9999  CDW
  |
  |           ------------- upgraded and renamed for 3.1 -----------
  |
  |           29Nov06 S0391423 Protect RMMXG from equals sign    CDW
  |           30Nov06 S0375339 Do not ABEND when on spawner      CDW
  |           21May07 S0405883 Global macro vars for generics    CDW
  |           01Oct09 S0365425 3.2 uses DUPLIB macro variable
  |                            instead of ADMIN. It points to
  |                            the "chosen one" staging library  CDW
  |           11Jan10 S0452584 Remove quotes for messaging macro CDW
  |           11Jan10 S0496289 Quote the sysprocessname variable CDW
  |           27Jan10 S0639651 Shorten some lines under 80 bytes CDW
  |           10Feb10 S0569529 Change CPCOLECT to ADAPTER and
  |                            remove unnecessary CPTOOLNM       CDW
  |           19Feb10 S0631019 For EREP, translate IDVAR into
  |                            characters for use in RMMSG       CDW
  |           29Jun10 S0363544 Use hash object instead of arrays CDW
  |           23Jul10 S0686401 Store SYSHOST while hash objects
  |                            initialize                        CDW
  |           07Oct11 S0750268 Improve performance when hundreds
  |                            of machines are processed, by
  |                            suppressing log messages and
  |                            writing to an audit table         CDW
  |           13Apr12 S0829745 Correct bad message key           CDW
  |           26Mar13 S0628462 Exempt potential duplicates from
  |                            being discarded if their
  |                            timestamps exactly match the end
  |                            of an existing range and appear
  |                            first in the input SMF file       CDW
  |           26Mar13 S0932137 Allow IDVAR variable to be longer
  |                            than 8 bytes                      CDW
  |
  | Notes:    This macro should be executed by the SAS program which
  |           reads the raw input data.  For MVS users processing
  |           their SMF data with a toolname of MXG, this macro will
  |           be executed automatically by the %CMPROCES macro when
  |           the DUPMODE= parameter is specified.  For users of the
  |           generic collector, this macro must be executed by the
  |           code which is reading the raw data and producing the
  |           staged data.  This means it must be coded at a point
  |           where the nominated IDVAR and TIMESTMP variables have
  |           been read and before any output processing occurs.  This
  |           macro will then test the timestamp/idvar combination
  |           against the datetime ranges built by the previous
  |           process execution. Any input data that have timestamps
  |           within the ranges of the data that have been collected
  |           before are discarded if DUPMODE=DISCARD.  A count of
  |           each IDVAR's input records and duplicates is kept so
  |           that a report can be produced at end-of-file. A
  |           temporary input filtering control dataset is
  |           created/appended to which contains the datetime range
  |           information for each system/machine read. This is then
  |           used by the next invocation of the RMDUPUPD macro to
  |           update the permanent control dataset.
  |
  |           The RMDUPMXG macro is useful when installing the input
  |           filtering macros into the code used for generic
  |           collector data. It can be used to make all the changes
  |           to MXG code. Please review that macro for further
  |           details.
  |
  | End
  +-------------------------------------------------------------------*/

%global CPDUP_UP;      /* Flag to signify duplication check is        */
%global CPMODE;        /* active. It is used by RMDUPUPD to ensure    */
                       /* that RMDUPCHK has been executed.            */
%let CPDUP_UP = Y;
%let FORCE = %qupcase(&FORCE);
%let TERM = %qupcase(&TERM);
%let REPORT = %qupcase(&REPORT);
%if &REPORT = %str() %then %let REPORT=YES;
length _new_old_idval $3;
_new_old_idval=_new_old_idval;
_rec_cnt=_rec_cnt;
_dup_cnt=_dup_cnt;
_dup_warning=_dup_warning;
_oldend=_oldend;
label _rec_cnt='New records processed'
      _dup_cnt='Duplicate records discarded'
      _new_old_idval="New or Old &IDVAR";

  /* Cannot use cpmode here to determine the value for abort. A cpmode*/
  /* of BATCH can be run in foreground and we do not want to abort the*/
  /* SAS session.                                                     */

%global CPABONXT;   /* Flag to abort on next datastep iteration.      */
%global CPOPSYS;    /* Operating system.                              */
%global CP_NMSG;    /* Level of messaging.                            */

%if not %symexist(ADAPTER) %then %global ADAPTER;
%if not %symexist(DUPLIB) %then %global DUPLIB;

%if &CPABONXT = YES %then %let abort = X;
%else %if &CPOPSYS = MVS and &SYSENV = BACK and
      %qupcase(&sysprocessname) ne %qupcase(OBJECT SERVER) %then
      %let abort = Y;
%else %let abort = N;

%global CPDUPMDE;   /* YES/NO                                         */
%global CPDUPFRC;   /* DUPMODE=FORCE,                                 */
%global CPDUPTRM;   /* DUPMODE=TERM, CPDUPFOR required for external   */
%global CPDUPFOR;   /* reference to FORCE=YES/NO flag.                */
%global CPDUPDBG;
%global CPDUP_ALLOW_EQ;   /* exempt exact timestamps from discard     */
%global CPDUP_ALLOW_MSG;  /* print msgs about exempted records        */

%if &ADAPTER = SMF and &CPDUPMDE ne YES %then
   %do;

   /*
   ERROR: Input filtering must be requested using the DUPMODE parameter
          of the Staging Transformation for execution of the RMDUPCHK
          macro to be successful.
          To preserve the integrity of your IT DataMart, SAS will
          terminate with a U1001 abend.
   */

   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                    dupmode_not_requested_logError,noquote)');
   _msgRc = resolve('%rmLogErr(ITMS,%quote(' || _msgtxt ||'))');


   %if &ABORT = Y %then abort abend 1001 %str(;);
   %else
      %do;
      call symput("CP&SOURCE.DSN",'N');
      call symput("CP&SOURCE.SET",'N');
      abort cancel;
      %end;
   run;
   %goto MEXIT;
   %end;

%if &RANGES ne %str() %then
   %do;

   %* ERROR: The RANGES parm is no longer supported.;

   %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
               dupchk_ranges_not_supported_logInfo,
               noquote));

   %end;

%if &SYSTEMS ne %str() %then
   %do;

   %* ERROR: The SYSTEMS parm is no longer supported.;

   %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
               dupchk_systems_not_supported_logInfo,
               noquote));

   %end;

%let parmerr = 0;

%if &SOURCE = %str() %then
   %do;

   %* ERROR: The SOURCE parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,SOURCE));

   %let parmerr = 1;
   %end;

%else %if %length(&SOURCE) ne 3 %then
   %do;

   %* ERROR: The RMDUPCHK SOURCE parm must be 3 characters long.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
                           invalid_src_parm_logError,
                           noquote));

   %let parmerr = 1;
   %end;

%else %if %index(ABCDEFGHIJKLMNOPQRSTUVWXYZ,
                 %upcase(%substr(&SOURCE,1,1))) = 0
   %then %do;

   %* ERROR: The RMDUPCHK SOURCE parm must start with a letter.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
                           invalid_src_parm_letter_logError,
                           noquote));

   %let parmerr = 1;
   %end;

%if &IDVAR = %str() %then
   %do;

   %* ERROR: The IDVAR parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,IDVAR));

   %let parmerr = 1;
   %end;

%else %if %length(&IDVAR) gt 32 %then
   %do;

   %* ERROR: The RMDUPCHK IDVAR parm must be a valid variable name.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
                           invalid_varnm_parm_logError,
                           noquote,IDVAR));

   %let parmerr = 1;
   %end;

%if &TIMESTMP = %str() %then
   %do;

   %* ERROR: The TIMESTMP parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,TIMESTMP));

   %let parmerr = 1;
   %end;

%else %if %length(&TIMESTMP) gt 8 %then
   %do;

   %* ERROR: The RMDUPCHK TIMESTMP parm must be a valid variable name.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
                           invalid_varnm_parm_logError,
                           noquote,TIMESTMP));
   %let parmerr = 1;
   %end;

%if &ENDFILE = %str() %then
   %do;

   %* ERROR: The ENDFILE parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,ENDFILE));

   %let parmerr = 1;
   %end;

%else %if %length(&ENDFILE) gt 8 %then
   %do;

   %* ERROR: The RMDUPCHK ENDFILE parm must be a valid variable name.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
                           invalid_varnm_parm_logError,
                           noquote,ENDFILE));
   %let parmerr = 1;
   %end;

%if &INT = %str() %then
   %do;

   %* ERROR: The INT parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,INT));

   %let parmerr = 1;
   %end;

%else %if %sysfunc(translate(&INT,nnnnnnnnnn,1234567890)) ne %str(nn:nn)
      and %sysfunc(translate(&INT,nnnnnnnnnn,1234567890)) ne %str(n:nn)
   %then %do;

   %* ERROR: The RMDUPCHK INT parm must be of the form HH:MM.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_HHMM_logError,
               noquote));

   %let parmerr = 1;
   %end;

%if &KEEP = %str() %then
   %do;

   %* ERROR: The KEEP parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,KEEP));


   %let parmerr = 1;
   %end;

%else %if %sysfunc(translate(&KEEP,nnnnnnnnnn,1234567890)) ne
   %substr(nnnnnnnnnn,1,%length(&KEEP)) %then
   %do;

     %* ERROR: The RMDUPCHK KEEP parm must be numeric.;

     %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_numeric_logError,
               noquote,KEEP));

     %let parmerr = 1;
   %end;

%if &FORCE = %str() %then
   %do;

   %* ERROR: The FORCE parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,FORCE));

   %let parmerr = 1;
   %end;

%if &TERM = %str() %then
   %do;

   %* ERROR: The TERM parm has not been supplied to RMDUPCHK.;

   %rmLogErr(ITMS,%rmmsg(sashelp.itms_msg,
               parm_not_supplied_logError,
               noquote,TERM));

   %let parmerr = 1;
   %end;

%if &REPORT ne YES and &REPORT ne NO %then
   %do;

   /* WARNING: The RMDUPCHK REPORT parm has an invalid value of &REPORT */
   /*          specified. It should be YES or NO. It is reset to YES.   */

   %rmLogWrn(ITMS,%rmmsg(sashelp.itms_msg,
                         dupchk_invalid_parm2_logWarn,
                         noquote,REPORT,&REPORT));
   %let report = YES;
   %end;

%if &parmerr = 1 %then
   %do;
   /*
     ERROR: Due to the previous error message and to preserve the
            integrity of your IT DataMart, SAS will terminate with
            a U1002 abend.
    */
   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                           previous_error_abend_logError,noquote)');
   _msgRc = resolve('%rmLogErr(ITMS,%quote(' || _msgtxt ||'))');

   %if &ABORT = Y %then abort abend 1002 %str(;);
   %else
      %do;
      call symput("CP&SOURCE.DSN",'N');
      call symput("CP&SOURCE.SET",'N');
      abort cancel;
      %end;
   %goto MEXIT;
   %end;

%global CPDUPEOF;        /* Flag to signify EOF reached on input      */
%global CP&SOURCE.DSN;   /* Flag used by RMDUPDSN to ensure that its  */
                         /* code only executes once per datastep. It  */
                         /* is set on the data stmt and unset at EOF. */

%if &DUPLIB =  %then %let DUPLIB=ADMIN;
%global AUDITDSN;
%if (&AUDITDSN = ) %then
   %let AUDITDSN = %upcase(&DUPLIB..&SOURCE.AUDIT);
%let CPDUPEOF = N;
%let CPBL5 = %str(     );
%let CPBL8 = %str(        );

%let CPADMXST = %sysfunc(exist(&DUPLIB..&SOURCE.cntrl));

%if &FORCE ne YES and &FORCE ne NO %then
   %do;

   %* WARNING: The RMDUPCHK FORCE parm has an invalid value of &FORCE;
   %*          specified. It should be YES or NO. It is reset to NO.;

     %rmLogWrn(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_invalid_parm_logWarn,
                            noquote,FORCE,&FORCE));

     %let force = NO;
   %end;

%global FRCTEXT;
%if &CPDUPFRC = YES %then %let FRCTEXT = %STR(DUPMODE=FORCE);
%if &FORCE = YES %then %let FRCTEXT = %STR(FORCE=YES);

%if &CPDUPFRC = NO and &FORCE = YES %then
   %do;

     %* Note: The RMDUPCHK FORCE parm of YES overrides the Staging ;
     %*       Transformation DUPMODE= parm. Duplicate input data ;
     %*       will be allowed.;

     %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_force_override_logInfo,noquote));

   %end;

%if &CPDUPFRC = YES and &FORCE = NO %then
   %do;

     %* Note: The Staging Transformation DUPMODE parm of FORCE ;
     %*       overrides the RMDUPCHK FORCE=NO parm. Duplicate ;
     %*       input data will be allowed.;

     %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_dupmode_override_logInfo,noquote));

   %let force = YES;
   %end;

%let CPDUPFOR = &FORCE;

%if &TERM ne YES and &TERM ne NO %then
   %do;

     %* WARNING: The RMDUPCHK TERM parm has an invalid value of &TERM;
     %*          specified. It should be YES or NO. It is reset to NO.;

     %rmLogWrn(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_invalid_parm_logWarn,
                            noquote,TERM,&TERM));

   %let term = NO;
   %end;

%global TRMTEXT;
%if &CPDUPTRM = YES %then %let TRMTEXT = %STR(DUPMODE=TERMINATE);
%if &TERM = YES %then %let TRMTEXT = %STR(TERM=YES);

%if &CPDUPTRM = NO and &TERM = YES %then
   %do;

     %* Note: The RMDUPCHK TERM parm of YES overrides the Staging;
     %*       Transformation DUPMODE= parm. SAS will stop if ;
     %*       duplicate input data are detected.;

     %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_term_override_logInfo,noquote));

   %end;

%if &CPDUPTRM = YES and &TERM = NO %then
   %do;

     %* Note: The Staging Transformation DUPMODE parm of TERMINATE;
     %*       overrides the RMDUPCHK TERM=NO parm. SAS will stop;
     %*       if duplicate input data are detected.;

     %rmLogInf(ITMS,%rmmsg(sashelp.itms_msg,
                            dupchk_dupmode_term_logInfo,noquote));

     %let term = YES;

   %end;

%global CP&SOURCE.SET; /* Flag used to ensure that this macro's setup */
                       /* code and datastep compile statements only   */
                       /* execute once. It's set here and unset at EOF*/

%if &&&CP&SOURCE.SET = Y %then %goto initdone;
%let CP&SOURCE.SET = Y;

 /*------------------------------------------------------------------+
  |                                                                  |
  |  Set up data step variables                                      |
  |                                                                  |
  +------------------------------------------------------------------*/

length _cpinit _cpdel 4;

retain _int "&INT"t _cpinit 0;

format _start _end datetime21.2;

drop _int _cpinit _cpgot _cpdel _chkdone _cptest2;

                                      /* Avoid error msg when an      */
if &IDVAR = ' ' then &IDVAR=&IDVAR;   /* incorrect idvar is given.    */
_eofonly = _eofonly;                  /* Avoid uninitialised msgs.    */

%if &CPABONXT = YES %then
   %do;
   retain _doabort 0;
   drop _doabort;
   %end;

goto skip_hash;

declare_hash:

   /* Hash object for new information on data being read */

   declare hash current_array(multidata: 'y', ordered: 'ascending',
                              hashexp: 8);
   declare hiter current_iter('current_array');
   _rc = current_array.defineKey("&IDVAR");
   _rc = current_array.defineData("&IDVAR", '_start', '_end', '_cnt');
   _rc = current_array.defineDone();

   /* Hash object for unique machine names being read    */

   declare hash current_idvals(ordered: 'ascending', hashexp: 8);
   declare hiter idvals_iter('current_idvals');
   _rc = current_idvals.defineKey("&IDVAR");
   _rc = current_idvals.defineData("&IDVAR", '_new_idval','_rec_cnt',
                                   '_first_rec');
   _rc = current_idvals.defineDone();
   call missing(_new_idval, _rec_cnt, _first_rec);
   retain _first_rec;

   %if &CPADMXST = 1 %then
      %do;

      format _oldstart _oldend datetime21.2;
      drop _oldstart _oldend;

      /* Hash object and iterator for existing control information    */
      /* from dataset - one entry per range per machine               */

      declare hash dupcntrl(dataset: "&DUPLIB..&SOURCE.CNTRL
                           (rename=(start=_oldstart end=_oldend))",
                   multidata: 'y', ordered: 'ascending', hashexp: 8);
      declare hiter dupiter('dupcntrl');

      /* Hash object and iterator for record counts for unique        */
      /* machine names from the control dataset                       */

      declare hash rec_count(ordered: 'ascending', hashexp: 2);
      declare hiter reccnt_iter('rec_count');


      _rc = dupcntrl.defineKey("&IDVAR");
      _rc = dupcntrl.defineData("&IDVAR", '_oldstart', '_oldend');
      _rc = dupcntrl.defineDone();
      call missing(_oldstart, _oldend);

      /* Hash object and iterator for record counts for unique        */
      /* machine names and initialize to missing                      */

      _rc = rec_count.defineKey("&IDVAR");
      _rc = rec_count.defineData("&IDVAR", '_rec_cnt', '_dup_cnt');
      _rc = rec_count.defineDone();
      call missing(_rec_cnt, _dup_cnt);

      _rc1 = dupiter.first();
      do while(_rc1 = 0);
         _rc2 = rec_count.find();
         if _rc2 ne 0 then _rc2 = rec_count.add();
         _rc1 = dupiter.next();
         end;

      %end;

   %else %if &cp_nmsg ge 1 %then
      %do;

      /*
         Note: The input filtering control data set
               &DUPLIB.&SOURCE.CNTRL, does not exist. It
               will be created.
      */

      _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                              dupchk_cntrl_created_logInfo,
                noquote,' || "%upcase(&DUPLIB..&SOURCE.CNTRL)" || ')');

      _msgRc = resolve('%rmLogInf(ITMS,%quote(' ||
                       _msgtxt || '))');

      %end;

   return;

skip_hash:

%INITDONE:

if not _eofonly then do;

 /*------------------------------------------------------------------+
  |                                                                  |
  | _CHKDONE is a flag set when the dup checking code has run.       |
  | Checking this variable ensures that the code is not run twice    |
  | even if the RMDUPCHK macro is specified twice.                   |
  |                                                                  |
  +------------------------------------------------------------------*/
   if _eofonly = 0 then _chkdone = 0;
   if not _chkdone then
      do;

 /*------------------------------------------------------------------+
  |                                                                  |
  |  Initialise system/machine name and datetime range arrays from   |
  |  the control data.                                               |
  |                                                                  |
  +------------------------------------------------------------------*/

      if not _cpinit then
         do;

         _cpinit = 1;

         _idvar = &IDVAR;

         link declare_hash;

         &IDVAR = _idvar;

         if &ENDFILE not in (0,1) then
            do;

            /*
            ERROR: The SAS variable, &ENDFILE, specified by the ENDFILE
               parm, is not in use as an end-of-file indicator.
               The ENDFILE parm must match the value of the
               END= keyword on the INFILE statement being used to
               read the external source data. To preserve the integrity
               of your IT DataMart, SAS will terminate with a U1003
               abend.
            */
            _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                      dupchk_end_match_logError,
                                      noquote,' || "&ENDFILE," ||')');

            _msgRc = resolve('%rmLogErr(ITMS,%quote('|| _msgtxt ||'))');

            %if &ABORT = X %then
               %do;
               _doabort = 1;
               return;
               %end;
            %else %if &ABORT = Y %then abort abend 1003 %str(;);
            %else
               %do;
               call symput("CP&SOURCE.DSN",'N');
               call symput("CP&SOURCE.SET",'N');
               abort cancel;
               %end;
            end;

         end; /* if not _cpinit then do */

 /*------------------------------------------------------------------+
  |                                                                  |
  | Check the incoming data. If this is for a system/machine that we |
  | have not seen already, add it to the array.                      |
  |                                                                  |
  +------------------------------------------------------------------*/

      _rc1 = current_idvals.find();
      if _rc1 = 0 then
         do;
         _rec_cnt + 1;
         _rc1 = current_idvals.replace();
         end;
      else
         do;
         %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:1 new idvalue '
             &IDVAR= %str(;) ;
         _new_idval = 1;
         _first_rec = 1;
         _rec_cnt = 1;
         _rc1 = current_idvals.add();
         end;

 /*------------------------------------------------------------------+
  |                                                                  |
  | Check the incoming data against the existing datetime ranges     |
  | to see if data needs to be thrown away.                          |
  |                                                                  |
  +------------------------------------------------------------------*/

      _cpdel = 0;

%if &CPADMXST = 1 %then
      %do;
      _rc1 = dupcntrl.find();
      if _rc1 = 0 then
         do;
         _rc2 = current_idvals.find();
         if _new_idval = 1 then
            do;
            _new_idval = 0;
            _rc2 = current_idvals.replace();
            end;
         _rc2 = rec_count.find();
         _rec_cnt + 1;
         _rc2 = rec_count.replace();
         end;

      %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:2 Is data in hash? '
         _rc1= &IDVAR= &TIMESTMP= %str(;) ;
      _cpgot = 0;
      do while(_rc1 = 0 and _cpgot = 0);
         if _oldstart le &TIMESTMP le _oldend then
            do;
            if "&CPDUP_ALLOW_EQ" = "YES" and &TIMESTMP = _oldend and
               _first_rec

%if &ADAPTER = SMF %then
                  and id not in(2,3);

               then
               do;
               if "&CPDUP_ALLOW_MSG" = "YES" then
                  put 'Potential duplicate has been processed '
                  _n_= system= id= smftime=;
               end;
            else
               do;
               _cpgot = 1;
               if 1
%if &ADAPTER = SMF %then
                  and id not in(2,3);
                  then
                  do;
                  _rc3 = current_idvals.find();
                  _first_rec = 0;
                  _rc3 = current_idvals.replace();
                  end;
            %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:2 Yes - ' &TIMESTMP
               'found in range ' _rc1= _oldstart= _oldend= %str(;) ;
               leave;
               end;
            end;
         else
            do;
            if 1
%if &ADAPTER = SMF %then
            and id not in(2,3);
               then
               do;
               _rc3 = current_idvals.find();
               _first_rec = 0;
               _rc3 = current_idvals.replace();
               end;
            end;
         _rc1 = dupcntrl.find_next();
         end;

      %if &CPDUPDBG ge 1 %then
         if not _cpgot then put 'DUP_DEBUG:2 No - ' &TIMESTMP
         'not found in any range ' %str(;) ;

      if _cpgot then
         do;
      %if &TERM = YES %then
         %do;

         /*
         ERROR: Duplicate input data detected for &IDVAR _syssnm
                You specified &TRMTEXT so SAS will terminate with a
                U1008 abend.
         */

         %if &ADAPTER = EREP %then
            %do;
         if notalnum(_syssnm) then __syssnm = put(trim(&IDVAR),hex32.);
                              else
            %end;
                                   __syssnm = &IDVAR;

         __syssnm = compress(__syssnm,"'");
         __syssnm = compress(__syssnm,'"');
         _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                   dupchk_dup_data_term_logError,
                                   noquote,' || "&IDVAR," || __syssnm ||
                                   %nrstr(",&TRMTEXT") || ')');

         _msgRc = resolve('%rmLogErr(ITMS,%quote(' ||
                             _msgtxt ||'))');

         %if &ABORT = X %then
            %do;
            _doabort = 1;
            return;
            %end;
         %else %if &ABORT = Y %then abort abend 1008 %str(;);
         %else
            %do;
            call symput("CP&SOURCE.DSN",'N');
            call symput("CP&SOURCE.SET",'N');
            abort cancel;
            %end;
         %end;

         _rc2 = rec_count.find();
         _dup_cnt + 1;
         _rc2 = rec_count.replace();
         _cpdel=1;
         end;

      %end;

 /*------------------------------------------------------------------+
  |                                                                  |
  |  Build an in-memory array of datetime ranges of the input data   |
  |  for each system/machine name, but only for observations for     |
  |  which the delete flag has not been set (_CPDEL).                |
  |                                                                  |
  +------------------------------------------------------------------*/

      if not _cpdel

%if &ADAPTER = SMF %then
         and id not in(2,3);

         then
         do;
         _cpgot = 0;
         _rc1 = current_array.find();
         %if &CPDUPDBG ge 1 %then put "DUP_DEBUG:3 Find &IDVAR " &IDVAR
            'in current_array ' _rc1= &TIMESTMP= %str(;) ;
         do while(_rc1 = 0 and _cpgot = 0);
            if _end le &TIMESTMP le _end+_int then
               do;
               _cpgot = 1;
               _end = &TIMESTMP;
               _cnt + 1;
               %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:3 ' &TIMESTMP
                  'extends range ' _start= _end= %str(;) ;
               leave;
               end;
            else if _start le &TIMESTMP le _end then
               do;
               _cpgot = 1;
               _cnt + 1;
               %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:3 ' &TIMESTMP
                  'fits in range ' _start= _end= %str(;) ;
               leave;
               end;
            else if _start ge &TIMESTMP ge _start-_int then
               do;
               _cpgot = 1;
               _start = &TIMESTMP;
               _cnt + 1;
               %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:3 ' &TIMESTMP
                  'prepends range ' _start= _end= %str(;) ;
               leave;
               end;
            _rc1 = current_array.find_next();
            end;

         %if &CPDUPDBG ge 1 %then
            if _cpgot = 1 then put 'DUP_DEBUG:3 Replacing '
            &IDVAR= _start= _end= _cnt= _rc1= %str(;) ;
         if _cpgot then _rc1 = current_array.replacedup();
         else
            do;
            /* Add new range */
            _start = &TIMESTMP;
            _end = &TIMESTMP;
            _cnt = 1;
            %if &CPDUPDBG ge 1 %then put 'DUP_DEBUG:3 ' &TIMESTMP
               'starts a new range ' _start= _end= %str(;) ;
            _rc1 = current_array.add();
            end;
         end; /* if not _cpdel then do */

      _chkdone = 1;
      end; /* if not _chkdone then do */

   end; /* if not _eofonly then do */


 /*------------------------------------------------------------------+
  |                                                                  |
  | End of file processing.  Build a report showing which IDVAR      |
  | values have produced the data and, for those IDVAR values that   |
  | have been read on previous executions, provide a record count.   |
  | Also provide warnings if duplicate input data are to be          |
  | deleted.                                                         |
  |                                                                  |
  +------------------------------------------------------------------*/

if (_eofonly or &ENDFILE) and (not _eofdone) then
   do;
   _idvar = &IDVAR;

   /*
      +----------------------------------------------------------------+
      | SAS IT Resource Management input data duplication check report
      | ==============================================================
      |
   */
   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                             dupchk_report_break_logInfo,
                             noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                             dupchk_check_report_logInfo,
                             noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                             dupchk_report_separator_logInfo,
                             noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                             dupchk_report_blank_logInfo,
                             noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   %if &REPORT=NO %then
      %do;
      /* Report suppressed. Audit trail can be found in table */

      _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                dupchk_report_auditdsn_logInfo,
                                noquote, ' || "&AUDITDSN" || ')');

      _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

      %end;

%if &CPADMXST = 1 %then
   %do;

   _rc1 = reccnt_iter.first();
   do while(_rc1 = 0);
      if _dup_cnt ge 1 then _dup_warning+1;
      %if &REPORT=NO %then
         %do;
         /* Write report record to audit table */
         _new_old_idval = 'Old';
         output &AUDITDSN;
         %end;
      %else
         %do;
         /* Write report record to SAS log */
         %if &CPDUPDBG ge 1 %then
            put "DUP_DEBUG:4 EOF report for &IDVAR " &IDVAR %str(;) ;

         %if &ADAPTER = EREP %then
            %do;
            if notalnum(_syssnm) then __syssnm = put(trim(&IDVAR),hex32.);
                                 else
            %end;
                                      __syssnm = &IDVAR;

         __syssnm=compress(__syssnm,"'");
         __syssnm=compress(__syssnm,'"');

         if _rec_cnt gt 0 then
            do;
            %if &CPDUPDBG ge 1 %then
               put "DUP_DEBUG:4 EOF report for &IDVAR " &IDVAR
               'New records processed ' _dup_cnt= _rec_cnt= %str(;) ;
            if _dup_cnt le 0 then
               do;

               /*
               Note: All rec_cnt input records for &IDVAR  _syssnm
                     will be added.
               */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                 dupchk_report_nodel_logInfo,
                                 noquote,'
                                 || put(_rec_cnt,8.) ||
                                 ",&IDVAR,"
                                 || __syssnm ||')');

               _msgRc = resolve('%rmLogInf(ITMS,%quote(' ||
                         _msgtxt ||'))');
               end;
            else
               do;
               /*
               WARNING: Duplicate input data detected for &IDVAR _syssnm
                        _rec_cnt incoming records" ;
               */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                 dupchk_report_del1_logWarn,
                                 noquote,' || "&IDVAR," || __syssnm ||')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' ||
                                _msgtxt ||'))');

               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                 dupchk_report_del2_logWarn,
                                 noquote,' || put(_rec_cnt,8.) ||')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' ||
                                _msgtxt ||'))');

               %if &FORCE ne YES %then
                  %do;
                  /*
                  |            _syssnmd incoming records were deleted.
                  |
                  */
                  _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                    dupchk_report_force_logWarn,
                                    noquote,' || put(_dup_cnt,8.) ||')');

                  _msgRc = resolve('%rmLogWrn(ITMS,%quote(' ||
                                   _msgtxt ||'))');

                  %end;
               %else
                  %do;
                  /*
                  |         of which _syssnmd would have been deleted.
                  |       You specified &FRCTEXT so they will be processed.
                  |
                  */
                  _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                    dupchk_report_del3_logWarn,
                                    noquote,' || put(_dup_cnt,8.) ||
                                    ',%quote('|| "&FRCTEXT" ||'))');

                  _msgRc = resolve('%rmLogWrn(ITMS,%quote(' ||
                                   _msgtxt ||'))');

                  %end;
               end;
            end;
         else
            do;/* _cpgot = 0 */
            /*
              | Note: No input data for &IDVAR  _syssnm were processed.
              |
            */
            %if &CPDUPDBG ge 1 %then
               put "DUP_DEBUG:4 EOF report for &IDVAR " &IDVAR
               'No new records processed' %str(;) ;
            _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                              dupchk_report_nodat_logInfo,
                              noquote,' || "&IDVAR," || __syssnm ||')');

            _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');
            end;
         %end;

      _rc1 = reccnt_iter.next();
      end;
   %end;

 /*------------------------------------------------------------------+
  |                                                                  |
  | Input data for systems/machines that have never been processed   |
  | before are reported here.                                        |
  |                                                                  |
  +------------------------------------------------------------------*/

   _rec_cnt=0;
   _dup_cnt=0;
   _rc1 = idvals_iter.first();
   do while(_rc1 = 0);
      if _new_idval = 1 then
         do;
         %if &REPORT=NO %then
            %do;
            /* Write report record to audit table */
            _new_old_idval = 'New';
            output &AUDITDSN;
            %end;
         %else
            %do;
            /*
            | Note: All input records for new &IDVAR __sysnm will be added.
            |
            */
            %if &ADAPTER = EREP %then
               %do;
               if notalnum(&IDVAR) then __sysnm = put(trim(&IDVAR),hex32.);
                                   else
               %end;
                                        __sysnm = &IDVAR;

            __sysnm=compress(__sysnm,"'");
            __sysnm=compress(__sysnm,'"');
            _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                              dupchk_report_new_logInfo,
                              noquote,'
                              || put(_rec_cnt,8.) ||
                              ",&IDVAR," || __sysnm ||')');

            _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');
            %end;
         end;

      _rc1 = idvals_iter.next();
      end;

   %if &REPORT=NO %then
      %do;
      if _dup_warning then
         do;
         _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                   dupchk_report_blank_logInfo,
                                   noquote)');

         _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

         /* At least one warning occurred - check AUDIT table */

         %if &FORCE ne YES %then
            %do;
            if _dup_warning = 1 then
               do;
               /* WARNING: Duplicates discarded for one IDVAR - check AUDITDSN. */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                         dupchk_report_auditdup1_logWarn,
                                         noquote, ' || "&IDVAR,&AUDITDSN" || ')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' || _msgtxt ||'))');
               end;
            else
               do;
               /* WARNING: Duplicates discarded for n IDVARs - check AUDITDSN.  */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                 dupchk_report_auditdupmany_logWarn,
                                 noquote,'
                                 || put(_dup_warning,8.) ||
                                 ",&IDVAR.s,&AUDITDSN" ||')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' || _msgtxt ||'))');
               end;
            %end;
         %else
            %do;
            if _dup_warning = 1 then
               do;
               /* WARNING: Duplicates forced in for one IDVAR - check AUDITDSN. */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                         dupchk_report_auditdupfrc1_logWarn,
                                         noquote, ' || "&IDVAR,&AUDITDSN" || ')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' || _msgtxt ||'))');
               end;
            else
               do;
               /* WARNING: Duplicates forced in for n IDVARs - check AUDITDSN.  */
               _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                                 dupchk_report_auditdupfrcmany_logWarn,
                                 noquote,'
                                 || put(_dup_warning,8.) ||
                                 ",&IDVAR.s,&AUDITDSN" ||')');

               _msgRc = resolve('%rmLogWrn(ITMS,%quote(' ||
                         _msgtxt ||'))');
               end;
            %end;
         end;
      %end;

   /*
     +----------------------------------------------------------------+
    */
   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                             dupchk_report_blank_logInfo,
                             noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   _msgtxt = resolve('%rmmsg(sashelp.itms_msg,
                           dupchk_report_break_logInfo,
                           noquote)');

   _msgRc = resolve('%rmLogInf(ITMS,%quote(' || _msgtxt ||'))');

   call symput("CPDUPEOF",'Y');
   call symput("CP&SOURCE.DSN",'N');
   call symput("CP&SOURCE.SET",'N');

 /*------------------------------------------------------------------+
  |                                                                  |
  | Update the working control dataset, appending to it if           |
  | necessary.                                                       |
  |                                                                  |
  +------------------------------------------------------------------*/

%if %sysfunc(exist(_lcpedup)) %then
   %do;
   if _cpnobsw ne 0 then
      do i = 1 to _cpnobsw;
      set _lcpedup nobs=_cpnobsw;
      output _lcpedup;
      end;
   %end;

   _cptest2 = &IDVAR;
   _cptest2 = repeat('*',199);
   _cpidlen = length(_cptest2);

%if &CPADMXST = 1 %then
   %do;
   _cptest1 = repeat('*',199);
   _cpidlen = max(length(_cptest1),length(_cptest2));
   drop _cptest1;
   %end;

   length _cpsourc $3
          _cpidvar $8
          _cpidval $200
          _cpstart _cpend _cpcnt 8;

   _cpsourc = "&SOURCE";
   _cpidvar = "&IDVAR";
   _cpint   = _int;
   _cpkeep  = &KEEP;

   format _cpstart _cpend datetime21.2;

   _rc1 = current_iter.first();
   do while(_rc1 = 0);
      if _cnt gt 0 then
         do;
         _cpidval = &IDVAR;
         _cpstart = _start;
         _cpend   = _end;
         _cpcnt   = _cnt;
         output _lcpedup;
         end;
      _rc1 = current_iter.next();
      end;

   _eofdone = 1;
   &IDVAR = _idvar;
   end; /* if (_eofonly or &ENDFILE) and (not _eofdone) then do */

 /*------------------------------------------------------------------+
  |                                                                  |
  |  Finally, if the delete flag has been set, that is, it is an     |
  |  observation that we don't want because it represents input      |
  |  data that has already been read, then delete it. We can't do    |
  |  this earlier cos the delete statement would stop further        |
  |  processing which could stop the end-of-file specific code from  |
  |  executing. _EOFONLY is only set by certain collectors that      |
  |  can't support delete statements.                                |
  |                                                                  |
  +------------------------------------------------------------------*/

if _cpdel and "&FORCE" ne "YES" then
   do;
   if _eofonly = . then delete;
   end;

%MEXIT:

%mend RMDUPCHK;
