/**********************************************************************/
/* Macro: rmsarstg                                                    */
/* ------------------------------------------------------------------ */
/* Description:                                                       */
/* rmsarstg contains parts of the SAR ETL process that may need to be */
/* modified as new versions and/or OSs are available.  This code can  */
/* be modified outside the java for quick and easy updates as needed. */
/* ------------------------------------------------------------------ */
/* History                                                            */
/* 30JUN2011 rmsarstg Production version 1.0                          */
/* ------------------------------------------------------------------ */
/* Copyright (c) 2011, 2012 SAS Institute, Inc Cary, NC USA           */
/* All rights reserved                                                */
/**********************************************************************/
%macro rmsarstg;

   %put %str(RMSARSTG version 1.0);

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_invalidLineCheck                              */
   /* Lines in the incoming SAR file to ignore.  New strings can be */
   /* added to the if condition.  _line contains the full line read */
   /* from the input file.  This macro is called very soon after    */
   /* a line is read from the input file.                           */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_invalidLineCheck;
      if (_line eq "" or
          upcase(strip(_line)) eq: "SYSTEM CONFIGURATION" or
          index(upcase(_line), "OPERATING SYSTEM IS RESTARTING")) then
         continue;
   %mend rmsarstg_invalidLineCheck;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_getMachineVersion                             */
   /* Get the machine and version from the header line. This        */
   /* information varies depending on the OS.                       */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_getMachineVersion;

      if (_upcaseOS eq "LINUX") then
      do;
         machine = scan(_line, 3, " ()");
         version = scan(_line, 2, " ");
         if upcase(scan(_line, -1, " ()")) eq 'CPU' then
         do;
            _dateString = scan(strip(_line), -4, ' ');
            date = input(strip(_dateString), anydtdte21.);
            chipArchitecture = scan(_line, 5, " ()");
            cpuNumber = scan(_line, 6, " ()");
         end;
         else
         do;
            chipArchitecture = "";
            cpuNumber = .;
         end;
      end;
      else
      do;
         machine = scan(_line, 2, " ");

         if (_upcaseOS eq "AIX") then
         do;
            _major = scan(_line, 3, " ");
            _minor = scan(_line, 4, " ");
            version = catx(".", _major, _minor);
         end;
         else if (_upcaseOS eq "DYNIX") then
            version = scan(_line, 4, " ");
         else
            version = scan(_line, 3, " ");
      end;

   %mend rmsarstg_getMachineVersion;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_validateOS                                    */
   /* Validate that this is a support OS.  Put out a warning        */
   /* message if the OS is nto in the list of valid OS's. Add       */
   /* new supported OS's to the if condition.                       */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_validateOS;

      if (_upcaseOS not in ("SUNOS",
                            "HP-UX",
                            "AIX",
                            "SCO_SV",
                            "LINUX",
                            "OSF1",
                            "DYNIX")) then
      do;
         _msgtxt = resolve('%rmmsg(sashelp.itms_msg,' !!
                           'SAR_fileNotProcessed_warning,noquote,' !!
                           strip(OS) !! ')');
         _msgRc = resolve('%rmLogWrn(ITMS,%quote(' || _msgtxt ||'))');
         %rcSetDS(4);
         return;
      end;

   %mend rmsarstg_validateOS;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_handleAverageSection                          */
   /* Handle sections that have "Average" instead of a time stamp   */
   /* at the beginning - set time to missing so that the records    */
   /* are deleted.                                                  */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_handleAverageSection;
      else if (strip(_timeString) eq: "Average") then
      do;
         time = .;
         validLine = 0;
      end;
   %mend rmsarstg_handleAverageSection;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_DoubleRowCheck                                */
   /* In general multiple row headers are not supported by the SAR  */
   /* adapter, but there are several cases where it is supported.   */
   /* This macro defines the special cases.  Add an else if for     */
   /* other special cases.                                          */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_DoubleRowCheck;
      else if (upcase(value{1}) eq '%AGE') then
      do;
         numValues + 1;
         do i = 2 to numValues;
            header{1, i} = catx("_", header{1, i}, upcase(value{i-1}));
         end;
         validValue = 0;
         numHeader = 1;
         valueRow = 1;
         headerFlag = 0;
         badSection = 0;
      end;
   %mend rmsarstg_DoubleRowCheck;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_ColumnHeaderProcessing                        */
   /* This macro handles any special processing of column headers.  */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_ColumnHeaderProcessing;
      if (index(header{valueRow, 1}, "-SZ")) then
      do;
         do i = 1 to numValues;
            if (header{valueRow, i} eq "OV") then
               header{valueRow, i} = strip(header{valueRow, i-1}) !!
                                     "-OV";
         end;
      end;

      if (header{valueRow, 1} eq "SML_MEM" and
          header{valueRow, 4} eq "LG_MEM" and
          header{valueRow, 7} eq "OVSZ_ALLOC") then
      do;
         header{valueRow, 2} = "SML_MEM_ALLOC";
         header{valueRow, 3} = "SML_MEM_FAIL";
         header{valueRow, 5} = "LG_MEM_ALLOC";
         header{valueRow, 6} = "LG_MEM_FAIL";
         header{valueRow, 8} = "OVSZ_FAIL";
      end;
      if (header{valueRow, 1} = "TPS") then
         header{valueRow, 1} = "TOTTPS";

      if (header{valueRow, 5} eq "PGFREE/S" and
          header{valueRow, 6} eq "PGSCANK/S") then
         header{valueRow, 5} = "PGFREE/S-ZLINUX";

   %mend rmsarstg_ColumnHeaderProcessing;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_MakeTempColumns                               */
   /* Creates any columns that are needed in calculations but are   */
   /* not being kept in the output tables.                          */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_MakeTempColumns;

      else if ((header{valueRow, i} eq '%IDLE' and not index(ignoreHeaders, ':%IDLE:'))) then
         IDLE = input(value{i}, 10.);

   %mend rmsarstg_MakeTempColumns;


   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_ValueProcessing                               */
   /* This macro handles any special processing of values.          */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_ValueProcessing;

      if (header{valueRow, 1} in ("DEVICE", "DEV")) then
      do;
         if (value{1} in (' ', '<<State','change>>')) then
            delete;
         if indexc(value{1}, ',') then
         do;
            PARTITN = scan(value{1}, 2, ',');
            DEVICE = scan(value{1}, 1, ',');
         end;
         else
         do;
            PARTITN = "";
            DEVICE = value{1};
         end;
      end;

      else if (header{valueRow, 1} eq "CPU") then
      do;
         CPUID = input(value{1}, ?? 10.);
         if (CPUID eq .) then
            CPUID = 99999;
      end;

      else if (header{valueRow, 1} in ('%USR', '%USER')) then
      do;
         CPUID = 99999;
      end;

      else if (header{valueRow, 1} eq: "C_HI") then
      do;
         HIT = input(translate(value{3},'   ','()%'),8.);
      end;

      else
      do;
         do i = 1 to numValues;
            if (header{valueRow, i} eq "PROC-SZ") then
            do;
               PROCSZC = input(scan(value{i},1,' /'),8.);
               PROCSZT = input(scan(value{i},2,' /'),8.);
            end;
            else if (header{valueRow, i} eq "INOD-SZ") then
            do;
               INODSZC = input(scan(value{i},1,' /'),8.);
               INODSZT = input(scan(value{i},2,' /'),8.);
            end;
            else if (header{valueRow, i} eq "FILE-SZ") then
            do;
               FILESZC = input(scan(value{i},1,' /'),8.);
               FILESZT = input(scan(value{i},2,' /'),8.);
            end;
            else if (header{valueRow, i} eq "THRD-SZ") then
            do;
               THRDSZC = input(scan(value{i},1,' /'),8.);
               THRDSZT = input(scan(value{i},2,' /'),8.);
            end;
            else if (header{valueRow, i} eq "LOCK-SZ") then
            do;
               LOCKSZC = input(scan(value{i},1,' /'),8.);
               LOCKSZT = input(scan(value{i},2,' /'),8.);
            end;
         end;  /* do i - loop through headers and values */
      end;  /* else */

      ignoreHeaders = ":DEVICE:CPU:PROC-SZ:INOD-SZ:FILE-SZ:THRD-SZ:LOCK-SZ:C_HI*:";

   %mend rmsarstg_ValueProcessing;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_Calculations                                  */
   /* This macro handles any special processing of calculated       */
   /* values.                                                       */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_Calculations;

      /* to avoid warnings about missing columns */
      if (0) then
      do;
         CPUID = .;
         PARTITN = "";
      end;

      /* calculate CPUBUSY */
      if (CPUID ne .) then
      do;
         if (IDLE eq . and WIO eq .) then
            CPUBUSY = .;
         else
         do;
            CPUBUSY=sum(100,-IDLE,-WIO);
            if CPUBUSY < 0 or CPUBUSY > 100 then
               CPUBUSY = .;
         end;
      end;

   %mend rmsarstg_Calculations;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_KeepExtra                                     */
   /* This macro contains a list of columns that need to be kept    */
   /* in the temporary staging table so that where clauses don't    */
   /* fail                                                          */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_KeepExtra;

      PARTITN CPUID

   %mend rmsarstg_KeepExtra;

   /* ------------------------------------------------------------- */
   /* Macro: rmsarstg_SummaryWhere                                  */
   /* This macro has additional where clause information for        */
   /* specific staging tables. This is used when creating the       */
   /* summarized tables.                                            */
   /* ------------------------------------------------------------- */
   /* Parameters:                                                   */
   /* table - name of the staging table to check                    */
   /* ------------------------------------------------------------- */
   %macro rmsarstg_SummaryWhere(table);

      %if (&table eq SAR or
           &table eq SARCPU or
           &table eq SARFAS or
           &table eq SARQUE or
           &table eq SARSEM or
           &table eq SARSWP or
           &table eq SARSYS) %then
      %do;
         and (CPUID eq 99999 or CPUID eq .)
      %end;

      %else %if (&table eq SARCPUB or
                 &table eq SARFASB or
                 &table eq SARQUEB or
                 &table eq SARSEMB or
                 &table eq SARSWPB or
                 &table eq SARSYSB) %then
      %do;
         and (CPUID ne 99999 and CPUID ne .)
      %end;

      %else %if (&table eq SARDEV) %then
      %do;
         and PARTITN eq ""
      %end;

      %else %if (&table eq SARPAG or
                 &table eq SARPG1) %then
      %do;
         and OS ne "AIX"
      %end;

      %else %if (&table eq SARPG3) %then
      %do;
         and OS eq "AIX"
      %end;

   %mend rmsarstg_SummaryWhere;

%mend rmsarstg;
